#!/usr/bin/perl

# Minim Madness

# Copyright (c) 2008 Martin Mariusz Lester
#
# Permission is hereby granted, free of charge, to any person
# obtaining a copy of this software and associated documentation
# files (the "Software"), to deal in the Software without
# restriction, including without limitation the rights to use,
# copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the
# Software is furnished to do so, subject to the following
# conditions:
#
# The above copyright notice and this permission notice shall be
# included in all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
# EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES
# OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
# NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT
# HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY,
# WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
# FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR
# OTHER DEALINGS IN THE SOFTWARE.

use strict;
use warnings;

use SDL;
use SDL::App;
use SDL::Surface;
use SDL::Event;
use SDL::Color;
use SDL::Rect;
use SDL::Tool::Graphic;
use SDL::Mixer;

use Storable qw(dclone);
use IO::File;
# For pi.
use Math::Trig;

# Handy sub-routines.

sub round {
  return sprintf("%.0f", $_[0]) + 0;
}

# Plot within the area of the app window only.
sub safe_plot {
  my ($surface, $x, $y, $colour) = @_;
  return if ($x < 0 || $x >= $surface->width() || $y < 0 || $y >= $surface->height());
  $surface->pixel($x, $y, $colour);
}

# Draw a line.
sub draw_line {
  # Bresenham-style line-drawing algorithm.
  my ($surface, $colour, $x1, $y1, $x2, $y2) = @_;

  # Just in case we get passed non-integers.
  ($x1, $y1, $x2, $y2) = (int $x1, int $y1, int $x2, int $y2);

  my $dx = $x2 - $x1;
  my $dy = $y2 - $y1;

  # Catch the case where the start and end point are the same.
  if ($dx == 0 && $dy == 0) {
    $surface->pixel($x1, $y1, $colour);
    return;
  }

  if (abs($dx) < abs($dy)) {
    # Line is mostly vertical.
    my $stepy = $dy > 0 ? 1 : -1;
    my $stepx = $dx/abs($dy);
    do {
      safe_plot($surface, $x1, $y1, $colour);
      $x1 += $stepx;
      $y1 += $stepy;
    } while ($y1 != $y2);
  }
  else {
    # Line is mostly horizontal.
    my $stepx = $dx > 0 ? 1 : -1;
    my $stepy = $dy/abs($dx);
    do {
      safe_plot($surface, $x1, $y1, $colour);
      $x1 += $stepx;
      $y1 += $stepy;
    } while ($x1 != $x2);
  }
  
  # Plot final pixel.
  safe_plot($surface, $x2, $y2, $colour);
}

# Draw a circle.
sub draw_circle {
  my ($surface, $colour, $x0, $y0, $r) = @_;

  # Just in case we get passed non-integers.
  ($x0, $y0, $r) = (int $x0, int $y0, int $r);

  my $f = 1 - $r;
  my $dx = 0;
  my $dy = -2 * $r;
  my $x = 0;
  my $y = $r;
  
  safe_plot($surface, $x0, $y0 + $r, $colour);
  safe_plot($surface, $x0, $y0 - $r, $colour);
  safe_plot($surface, $x0 + $r, $y0, $colour);
  safe_plot($surface, $x0 - $r, $y0, $colour);
  
  while ($x < $y) {
    if ($f >= 0) {
      $y--;
      $dy += 2;
      $f += $dy;
    }
    $x++;
    $dx += 2;
    $f += $dx + 1;

    safe_plot($surface, $x0 + $x, $y0 + $y, $colour);
    safe_plot($surface, $x0 + $x, $y0 - $y, $colour);
    safe_plot($surface, $x0 - $x, $y0 + $y, $colour);
    safe_plot($surface, $x0 - $x, $y0 - $y, $colour);
    safe_plot($surface, $x0 + $y, $y0 + $x, $colour);
    safe_plot($surface, $x0 + $y, $y0 - $x, $colour);
    safe_plot($surface, $x0 - $y, $y0 + $x, $colour);
    safe_plot($surface, $x0 - $y, $y0 - $x, $colour);
  }

}

# Draw a semibrieve.
sub draw_semi {
  my ($surface, $colour, $x0, $y0, $r) = @_;

  # Vertical zoom.
  my $z = 0.8;

  # Just in case we get passed non-integers.
  ($x0, $y0, $r) = (int $x0, int $y0, int $r);

  my $f = 1 - $r;
  my $dx = 0;
  my $dy = -2 * $r;
  my $x = 0;
  my $y = $r;
  
  safe_plot($surface, $x0, $y0 + int($r*$z), $colour);
  safe_plot($surface, $x0, $y0 - int($r*$z), $colour);
  safe_plot($surface, $x0 + $r, $y0, $colour);
  safe_plot($surface, $x0 - $r, $y0, $colour);
  
  while ($x < $y) {
    if ($f >= 0) {
      $y--;
      $dy += 2;
      $f += $dy;
    }
    $x++;
    $dx += 2;
    $f += $dx + 1;

    safe_plot($surface, $x0 + $x, $y0 + int($y*$z), $colour);
    safe_plot($surface, $x0 + $x, $y0 - int($y*$z), $colour);
    safe_plot($surface, $x0 - $x, $y0 + int($y*$z), $colour);
    safe_plot($surface, $x0 - $x, $y0 - int($y*$z), $colour);
    safe_plot($surface, $x0 + $y, $y0 + int($x*$z), $colour);
    safe_plot($surface, $x0 + $y, $y0 - int($x*$z), $colour);
    safe_plot($surface, $x0 - $y, $y0 + int($x*$z), $colour);
    safe_plot($surface, $x0 - $y, $y0 - int($x*$z), $colour);
  }

}

# Draw a minim.
sub draw_minim {
  my ($surf, $colour, $x1, $y1, $x2, $y2, $x3, $y3, $r, $length, $angle) = @_;
  draw_line($surf, $colour, $x1, $y1, $x2, $y2);
  draw_circle($surf, $colour, $x3, $y3, $r);
}

# Find Euclidean distance between two points.
sub distance {
  my ($x1, $y1, $x2, $y2) = @_;
  my $dx = $x2 - $x1;
  my $dy = $y2 - $y1;
  return sqrt(($dx * $dx) + ($dy * $dy));
}

# Print a string on a surface in a certain position, using a pre-rendered font.
sub font_print {
  my ($font, $surf, $text, $x, $y) = @_;

  my $rect = $font->{rect};
  my $targ = $font->{targ};
  $targ->y($y);
  for my $char (split //, $text) {
    my $glyph = $font->{uc($char)};
    if (defined($glyph)) {
      $targ->x($x);
      $glyph->blit($rect, $surf, $targ);
    }
    $x += int(1.2 * $rect->width());
  }
}

# Draw a notice on-screen.
sub notice {
  my ($app, $font, $text, $framecol, $bgcol, $x_mid, $y_mid) = @_;
  $x_mid = 320 if (!defined($x_mid));
  $y_mid = 240 if (!defined($y_mid));

  my $font_width = $font->{width};
  my $font_height = $font->{height};
  my $font_x_gap = int(0.2 * $font_width);
  my $font_y_gap = int($font_height / 2);
  
  my @lines = split (/\n/, $text);
  my $text_length = 0;
  foreach my $line (@lines) {
    $text_length = length($line) if length($line) > $text_length;
  }
  
  my $text_width = ($text_length * ($font_width + $font_x_gap)) - $font_x_gap;
  my $notice_width = $text_width + (2 * $font_width);
  my $x = $x_mid - ($notice_width/2);
  my $notice_height = (((scalar @lines) * ($font_height + $font_y_gap)) - $font_y_gap) + (2 * $font_height);
  my $y = $y_mid - ($notice_height/2);

  my $frame = new SDL::Rect(-width => $notice_width, -height => $notice_height, -x => $x, -y => $y);
  my $bg = new SDL::Rect(-width => $notice_width-2, -height => $notice_height-2, -x => $x+1, -y => $y+1);
  $app->fill($frame, $framecol);
  $app->fill($bg, $bgcol);
  
  $y += $font_height;
  foreach my $line (@lines) {
    font_print($font, $app, $line, $x_mid - (((length($line) * ($font_width + $font_x_gap)) - $font_x_gap)/2), $y);
    $y = $y + $font_height + $font_y_gap;
  }
}


# One-off initialisation.


# Create the game window.

my $app = new SDL::App(
  -title => "Minim Madness",
  -width => 640, -height => 480,
  -depth => 16,
);

# Rectangle the size of app window.
my $app_rect = new SDL::Rect(
  -x => 0, -y => 0,
  -width => 640, -height => 480,
);

# Prepare to receive SDL mouse/key events.
my $event = new SDL::Event();

# Load sounds.
my $play = sub { };
eval {
  my $mixer = new SDL::Mixer();
  my @sounds = qw(bounce break timenear timeout clear exit click easy hard);
  my %sounds = (map {$_ => new SDL::Sound("$_.wav")} @sounds);

  $play = sub { $mixer->play_channel(-1, $sounds{$_[0]}, 0); };
};

my $nosound = 0;
if ($@) {
  print "Sound initialisation failed: $@";
  $nosound = 1;
}

# Set up colours.
my $red		= new SDL::Color(-r => 0xff, -g => 0x00, -b => 0x00);
my $green	= new SDL::Color(-r => 0x00, -g => 0xff, -b => 0x00);
my $blue	= new SDL::Color(-r => 0x00, -g => 0x00, -b => 0xff);
my $black	= new SDL::Color(-r => 0x00, -g => 0x00, -b => 0x00);
my $white	= new SDL::Color(-r => 0xff, -g => 0xff, -b => 0xff);
my $yellow	= new SDL::Color(-r => 0xff, -g => 0xff, -b => 0x00);
my $cyan	= new SDL::Color(-r => 0x00, -g => 0xff, -b => 0xff);
my $magenta	= new SDL::Color(-r => 0xff, -g => 0x00, -b => 0xff);
my $grey	= new SDL::Color(-r => 0x80, -g => 0x80, -b => 0x80);

# Simple vector font.
my %font_vec = (
  A => [[0, 1, 0.5, 0], [0.5, 0, 1, 1], [0.25, 0.5, 0.75, 0.5]],
  B => [[0, 0, 0, 1], [0, 0, 0.5, 0], [0, 0.5, 0.5, 0.5], [0, 1, 0.5, 1], [0.5, 0, 1, 0.25], [0.5, 0.5, 1, 0.25], [0.5, 0.5, 1, 0.75], [0.5, 1, 1, 0.75]],
  C => [[1, 0, 0, 0], [0, 0, 0, 1], [0, 1, 1, 1]],
  D => [[0, 0, 0, 1], [0, 0, 0.5, 0], [0.5, 1, 0, 1], [0.5, 0, 1, 0.5], [1, 0.5, 0.5, 1]],
  E => [[0, 0, 0, 1], [0, 0, 1, 0], [0, 0.5, 1, 0.5], [0, 1, 1, 1]],
  F => [[0, 0, 0, 1], [0, 0, 1, 0], [0, 0.5, 1, 0.5]],
  G => [[1, 0.25, 1, 0], [1, 0, 0, 0], [0, 0, 0, 1], [0, 1, 1, 1], [1, 1, 1, 0.5], [1, 0.5, 0.5, 0.5]],
  H => [[0, 0, 0, 1], [1, 0, 1, 1], [0, 0.5, 1, 0.5]],
  I => [[0, 0, 1, 0], [0, 1, 1, 1], [0.5, 0, 0.5, 1]],
  J => [[0, 0, 1, 0], [0.5, 0, 0.5, 1], [0.5, 1, 0, 1], [0, 1, 0, 0.5]],
  K => [[0, 0, 0, 1], [0, 0.5, 1, 0], [0, 0.5, 1, 1]],
  L => [[0, 0, 0, 1], [0, 1, 1, 1]],
  M => [[0, 0, 0, 1], [1, 0, 1, 1], [0, 0, 0.5, 0.5], [0.5, 0.5, 1, 0]],
  N => [[0, 0, 0, 1], [1, 0, 1, 1], [0, 0, 1, 1]],
  O => [[0, 0, 0, 1], [0, 1, 1, 1], [1, 1, 1, 0], [1, 0, 0, 0]],
  P => [[0, 0, 0, 1], [0, 0, 1, 0], [0, 0.5, 1, 0.5], [1, 0, 1, 0.5]],
  Q => [[0.25, 0, 0.75, 0], [0.75, 0, 1, 0.25], [1, 0.25, 1, 0.75], [1, 0.75, 0.75, 1], [0.75, 1, 0.25, 1], [0.25, 1, 0, 0.75], [0, 0.75, 0, 0.25], [0, 0.25, 0.25, 0], [0.5, 0.5, 1, 1]],
  R => [[0, 0, 0, 1], [0, 0, 1, 0], [0, 0.5, 1, 0.5], [1, 0, 1, 0.5], [0.5, 0.5, 1, 1]],
  S => [[1, 0, 0, 0], [0, 0, 0, 0.5], [0, 0.5, 1, 0.5], [1, 0.5, 1, 1], [1, 1, 0, 1]],
  T => [[0, 0, 1, 0], [0.5, 0, 0.5, 1]],
  U => [[0, 0, 0, 1], [0, 1, 1, 1], [1, 1, 1, 0]],
  V => [[0, 0, 0.5, 1], [0.5, 1, 1, 0]],
  W => [[0, 0, 0.25, 1], [0.25, 1, 0.5, 0.5], [0.5, 0.5, 0.75, 1], [0.75, 1, 1, 0]],
  X => [[0, 0, 1, 1], [0, 1, 1, 0]],
  Y => [[0, 0, 0.5, 0.5], [0.5, 0.5, 1, 0], [0.5, 1, 0.5, 0.5]],
  Z => [[0, 0, 1, 0], [1, 0, 0, 1], [0, 1, 1, 1]],
  0 => [[0, 0, 1, 0], [1, 0, 1, 1], [1, 1, 0, 1], [0, 1, 0, 0], [1, 0, 0, 1]],
  1 => [[0, 1, 1, 1], [0.5, 0, 0.5, 1], [0.5, 0, 0, 0.5]],
  2 => [[1, 1, 0, 1], [0, 1, 1, 0.5], [1, 0.5, 0.5, 0], [0.5, 0, 0, 0.5]],
  3 => [[0, 0, 1, 0], [1, 0, 1, 1], [1, 1, 0, 1], [0.5, 0.5, 1, 0.5]],
  4 => [[0.5, 1, 0.5, 0], [0.5, 0, 0, 0.5], [0, 0.5, 1, 0.5]],
  5 => [[1, 0, 0, 0], [0, 0, 0, 0.5], [0, 0.5, 1, 0.5], [1, 0.5, 1, 1], [1, 1, 0, 1]],
  6 => [[0, 0.5, 1, 0.5], [1, 0.5, 1, 1], [1, 1, 0, 1], [0, 1, 0, 0.5], [0, 0.5, 1, 0]],
  7 => [[0, 0, 1, 0], [1, 0, 0, 1]],
  8 => [[0, 0, 1, 0], [1, 0, 1, 1], [1, 1, 0, 1], [0, 1, 0, 0], [0, 0.5, 1, 0.5]],
  9 => [[0, 0, 1, 0], [1, 0, 1, 0.5], [1, 0.5, 0, 0.5], [0, 0.5, 0, 0], [1, 0.5, 0, 1]],
  "." => [[0.45, 1, 0.55, 1], [0.55, 1, 0.55, 0.9], [0.55, 0.9, 0.45, 0.9], [0.45, 0.9, 0.45, 1]],
  " " => [],
);

# Prepare the font in various colours.
my @fonts = ();
{

  foreach my $font_style ([$red, 20, 20], [$green, 20, 20], [$blue, 20, 20], [$white, 20, 20], [$white, 10, 10]) {
    my ($font_col, $font_width, $font_height) = @{$font_style};
    my $font_rect = new SDL::Rect (-width => $font_width, -height => $font_height, -x => 0, -y => 0);
    my $font_targ = new SDL::Rect (-width => $font_width, -height => $font_height, -x => 0, -y => 0);

    my %font_surf = ();
    $font_surf{rect} = $font_rect;
    $font_surf{targ} = $font_targ;
    $font_surf{width} = $font_width;
    $font_surf{height} = $font_height;

    foreach my $char (keys %font_vec) {
      my $vecs = $font_vec{$char};
      my $surf = new SDL::Surface(-width => $font_width, -height => $font_height);
      $surf->display_format();
      $surf->fill(0, $black);
      foreach my $vec (@{$vecs}) {
        my ($x1, $y1, $x2, $y2) = @{$vec};
        $x1 *= ($font_width - 1);
        $x2 *= ($font_width - 1);
        $y1 *= ($font_height - 1);
        $y2 *= ($font_height - 1);
        # Decent rounding is important here.
        draw_line($surf, $font_col, round($x1), round($y1), round($x2), round($y2));
      }
      $font_surf{$char} = $surf;
    }
    push (@fonts, {%font_surf});
  }
}

# Pre-render the semibreves in various colours.
my @semis = ();
my $semi_rad = 5;
my $semi_rect = new SDL::Rect(-x => 0, -y => 0, -width => 1+(2*$semi_rad), -height => 1+(2*$semi_rad));
my $semi_targ = new SDL::Rect(-x => 0, -y => 0, -width => 1+(2*$semi_rad), -height => 1+(2*$semi_rad));
foreach my $colour ($red, $green, $blue) {
  my $surf = new SDL::Surface(-width => 1+(2*$semi_rad), -height => 1+(2*$semi_rad), -flags => SDL_SRCCOLORKEY);
  $surf->display_format();
  $surf->set_color_key(SDL_SRCCOLORKEY, $black);
  $surf->fill(0, $black);
  draw_semi($surf, $colour, $semi_rad, $semi_rad, $semi_rad);
  push (@semis, $surf);
}

# Pre-render the target boxes.
my $box_rect = new SDL::Rect(-x => 0, -y => 0, -width => 20, -height => 20);
my $box_targ = new SDL::Rect(-x => 0, -y => 0, -width => 20, -height => 20);
my @boxes = ();
foreach my $colour ($red, $green, $blue) {
  my $surf = new SDL::Surface(-width => 20, -height => 20, -flags => SDL_SRCCOLORKEY);
  $surf->display_format();
  $surf->set_color_key(SDL_SRCCOLORKEY, $black);
  $surf->fill(0, $black);
  draw_line($surf, $colour, 0, 0, 19, 0);
  draw_line($surf, $colour, 19, 0, 19, 19);
  draw_line($surf, $colour, 19, 19, 0, 19);
  draw_line($surf, $colour, 0, 19, 0, 0);
  draw_line($surf, $colour, 0, 0, 19, 19);
  draw_line($surf, $colour, 19, 0, 0, 19);
  push (@boxes, $surf);
}

# Pre-render the quaver rests.
my $rest_rect = new SDL::Rect(-x => 0, -y => 0, -width => 20, -height => 20);
my $rest_targ = new SDL::Rect(-x => 0, -y => 0, -width => 20, -height => 20);
my $rest = new SDL::Surface(-width => 20, -height => 20, -flags => SDL_SRCCOLORKEY);
$rest->display_format();
$rest->set_color_key(SDL_SRCCOLORKEY, $black);
$rest->fill(0, $black);
draw_line($rest, $yellow, 5, 0, 10, 5);
draw_line($rest, $yellow, 10, 5, 15, 0);
draw_line($rest, $yellow, 15, 0, 10, 19);

# Prepare 3 surfaces: one for the background, one for minims, one for player drawings.
my $bg_surf = new SDL::Surface(-width => 640, -height => 480);
$bg_surf->display_format();
my $note_surf = new SDL::Surface(-width => 640, -height => 480, -flags => SDL_SRCCOLORKEY);
$note_surf->display_format();
$note_surf->set_color_key(SDL_SRCCOLORKEY, $black);
my $draw_surf = new SDL::Surface(-width => 640, -height => 480, -flags => SDL_SRCCOLORKEY);
$draw_surf->display_format();
$draw_surf->set_color_key(SDL_SRCCOLORKEY, $black);


# Load user levels if present.
my $corrupt = 0;
my $custom = 0;
my @game_levels = ();

{
  my @user_levels = ();
  my $source = new IO::File("levels", "<");
  last if !defined($source);

  my %colours = (red => 0, green => 1, blue => 2);
  my $n = qr{[-+]?\d+};
  my $f = qr{$n|$n\.\d*|\.\d+};

  my $linenum = 0;
  my $level = undef;

  while (!$source->eof()) {
    my $line = $source->getline();
    $linenum++;
    
    # Strip comments.
    if ($line =~ m/^(.*?)#/ ) {
      $line = $1;
    }
    
    # Ignore blank lines.
    next if ($line =~ m/^\s*$/ );

    # Strip leading and trailing space.
    if ($line =~ m/^(.*?)\s+$/ ) {
      $line = $1;
    }
    if ($line =~ m/^\s+(.*)$/ ) {
      $line = $1;
    }
    
    if ($line =~ m/^\s*level\s(.*)$/ ) {
      push (@user_levels, $level) if defined($level);
      $level = {
        name		=> $1,
        time		=> 100,
        targets		=> [0, 0, 0],
        exits		=> [],
        rests		=> [],
        minims		=> [],
        streams		=> [],
        text		=> [],
        demo_strokes	=> [],
      };
      next;
    }
    
    if ($line =~ m/^time\s+($n)$/ ) {
      $level->{time} = $1;
      next;
    }
    
    if ($line =~ m/^target\s+(\w+)\s+($n)$/ ) {
      my $col = $colours{$1};
      my $targ = $2;
      if (defined($col)) {
        $level->{targets}->[$col] = $targ;
        next;
      }
    }
    
    if ($line =~ m/^exit\s+(\w+)\s+($n)\s+($n)$/ ) {
      my $col = $colours{$1};
      if (defined($col)) {
        push (@{$level->{exits}}, [$2, $3, $col]);
        next;
      }
    }
    
    if ($line =~ m/^rest\s+($n)\s+($n)$/ ) {
      push (@{$level->{rests}}, [$1, $2]);
      next;
    }

    if ($line =~ m/^minim\s+($n)\s+($n)\s+($n)\s+($n)\s+($n)$/ ) {
      push (@{$level->{minims}}, [$1, $2, $3, $4, $5]);
      next;
    }

    if ($line =~ m/^stream\s+(\w+)\s+($n)\s+($n)\s+($f)\s+($f)\s+($n)\s+($n)\s+($n)$/ ) {
      my $col = $colours{$1};
      if (defined($col)) {
        push (@{$level->{streams}}, [$2, $3, $4+0.0, $5+0.0, $6, $7, $8, $col]);
        next;
      }
    }

    if ($line =~ m/^text\s+($n)\s+($n)\s+($n)\s(.*)$/ ) {
      if (($3 >= 0) && ($3 <= $#fonts)) {
        push (@{$level->{text}}, [$4, $1, $2, $3]);
      }
      next;
    }

    if ($line =~ m/^stroke\s+(.*)$/ ) {
      if ($1 eq "easy") {
        push (@{$level->{demo_strokes}}, "easy");
        next;
      }
      elsif ($1 eq "hard") {
        push (@{$level->{demo_strokes}}, "hard");
        next;
      }
      else {
        my @stroke = ($1 =~ m/($n)\s*/g );
        if ((scalar @stroke) % 3 == 0) {
          push (@{$level->{demo_strokes}}, [@stroke]);
          next;
        }
      }    
    }

    if ($line =~ m/^gravity\s+($f)\s+($f)$/ ) {
      $level->{gravity} = [$1+0.0, $2+0.0];
      next;
    }

    if ($line =~ m/^elasticity\s+($f)$/ ) {
      $level->{elasticity} = $1+0.0;
      next;    
    }    

    if ($line =~ m/^boost\s+($f)$/ ) {
      $level->{boost} = $1+0.0;
      next;    
    }    

    if ($line =~ m/^slow\s+($f)$/ ) {
      $level->{slow} = $1+0.0;
      next;    
    }    

    $corrupt = $linenum if (!$corrupt);

  }

  push (@user_levels, $level);
  
  $source->close();
  
  if (!$corrupt) {
    # Sort streams by starting frame.
    foreach my $level (@user_levels) {
      @{$level->{streams}} = sort {$a->[4] <=> $b->[4]} (@{$level->{streams}});
    }

    # Use the user-defined levels if there were no errors.
    @game_levels = @user_levels;
    $custom = @user_levels;
  }
}


# Demo level data.
my @demo_levels = (
  {
    name		=> "   BEST NOTION SOFTWARE",
    time		=> 30,
    targets		=> [0, 0, 0],
    exits		=> [],
    rests		=> [],
    minims		=> [[639, 390, 0, 390, 0],
                            [180, 80, 180, 120, 5], [220, 80, 220, 120, 5], [180, 80, 195, 95, 5], [220, 80, 205, 95, 5],
                            [240, 80, 280, 80, 5], [280, 120, 240, 120, 5], [260, 120, 260, 85, 5],
                            [300, 80, 300, 120, 5], [300, 80, 340, 120, 5], [340, 120, 340, 80, 5],
                            [360, 80, 400, 80, 5], [400, 120, 360, 120, 5], [380, 80, 380, 115, 5],
                            [420, 80, 420, 120, 5], [460, 80, 460, 120, 5], [420, 80, 435, 95, 5], [460, 80, 445, 95, 5]],
    demo_strokes	=> [[5, 120, 180, 10, 120, 140, 15, 140, 160, 20, 160, 140, 25, 160, 180,
                            30, 180, 180, 35, 200, 140, 40, 220, 180,
                            45, 240, 180, 50, 240, 140, 55, 260, 140, 60, 280, 160, 65, 260, 180,
                            70, 300, 180, 75, 300, 140, 80, 340, 180, 85, 340, 140,
                            90, 400, 140, 95, 360, 150, 100, 400, 160, 105, 360, 170, 110, 400, 180,
                            115, 460, 140, 120, 420, 153, 125, 460, 167, 130, 420, 180, 135, 460, 180,
                            140, 520, 140, 145, 480, 153, 150, 520, 167, 155, 480, 180, 160, 480, 180, 1000, 480, 180]],
    streams		=> [[320, 385, 2, -15, 0, 60, 26, 2], [320, 385, 4, -15, 0, 60, 24, 2], [320, 385, -2, -15, 0, 60, 27, 2], [320, 385, -4, -15, 0, 60, 23, 2],
                            [0, 240, 3, -15, 0, 60, 26, 0], [0, 240, 5, -15, 0, 60, 24, 0], [0, 240, 3, -10, 0, 60, 27, 0], [0, 240, 5, -10, 0, 60, 23, 0],
                            [639, 240, -3, -15, 0, 60, 26, 1], [639, 240, -5, -15, 0, 60, 24, 1], [639, 240, -3, -10, 0, 60, 27, 1], [639, 240, -5, -10, 0, 60, 23, 1]],
    text		=> [["PRESENTS", 226, 30, 3],
                            ["CLICK TO PLAY", 166, 400, 3],
                            ["PRESS S FOR INSTRUCTIONS", 34, 430, 3],
                            ["V1.0", 590, 0, 4], 
                            ($nosound ? ["FAILED TO START SOUND", 194, 200, 4] : ()),
                            ($custom ? ["$custom CUSTOM LEVELS LOADED", 195 - (6 * length($custom)), 215, 4] : ()),
                            ($corrupt ? ["ERROR IN LEVELS FILE ON LINE $corrupt", 147 - (6 * length($corrupt)), 215, 4] : ()) ],
  },

  {
    name		=> "AIM OF THE GAME",
    time		=> 30,
    targets		=> [0, 0, 0],
    exits		=> [[100, 360, 0], [360, 360, 1], [540, 360, 2]],
    rests		=> [],
    minims		=> [],
    streams		=> [[0, 150, 5, 0, 5, 25, 25, 0], [300, 380, 1.5, -20, 5, 25, 25, 1], [640, 150, -5, 0, 5, 25, 25, 2]],
    text		=> [["THE AIM OF THE GAME IS TO ", 0, 40, 3],
                            ["GUIDE SEMIBREVES TO EXITS ", 0, 70, 3],
                            ["OF THE SAME COLOUR        ", 0, 100, 3],
                            ["PRESS S TO SKIP TO NEXT   ", 0, 390, 3],
                            ["SCREEN OF INSTRUCTIONS    ", 0, 420, 3]],
  },

  {
    name		=> "AIM OF THE GAME",
    time		=> 20,
    targets		=> [0, 0, 0],
    exits		=> [[540, 400, 0]],
    rests		=> [],
    minims		=> [],
    demo_strokes	=> [[100, 380, 400, 110, 280, 400, 120, 180, 400, 130, 170, 375, 140, 170, 360, 150, 190, 360, 160, 190, 375, 170, 180, 400, 180, 180, 400],
                            [250, 0, 0, 250, 0, 0]],
    streams		=> [[0, 110, 10, 0, 0, 16, 25, 0]],
    text		=> [["YOU GUIDE THE SEMIBREVES  ", 0, 40, 3],
                            ["BY DRAWING MINIMS         ", 0, 70, 3]],
  },

  {
    name		=> "MINIMS AND SEMIBREVES",
    time		=> 20,
    targets		=> [0, 0, 0],
    exits		=> [],
    rests		=> [],
    minims		=> [[550, 200, 450, 200, 10], [250, 400, 150, 300, 13]],
    streams		=> [[640, 110, -10, 0, 5, 15, 25, 1]],
    text		=> [["SEMIBREVES BOUNCE OFF THE ", 0, 40, 3],
                            ["HEAD SIDE OF MINIMS       ", 0, 70, 3]],
  },

  {
    name		=> "MINIMS AND SEMIBREVES",
    time		=> 20,
    targets		=> [0, 0, 0],
    exits		=> [],
    rests		=> [],
    minims		=> [[100, 400, 100, 200, 20], [200, 400, 200, 200, 20], [300, 400, 300, 200, 20],
                            [400, 400, 400, 200, 20], [500, 400, 500, 200, 20], [600, 400, 600, 200, 20]],
    streams		=> [[0, 200, 20, 0, 5, 10, 25, 2], [0, 200, 40, 0, 11, 16, 25, 2]],
    text		=> [["SEMIBREVES DESTROY MINIMS ", 0, 40, 3],
                            ["AND ARE THEMSELVES        ", 0, 70, 3],
                            ["DESTROYED WHEN THEY HIT   ", 0, 100, 3],
                            ["THE OTHER SIDE            ", 0, 130, 3]],
  },

  {
    name		=> "MINIMS AND SEMIBREVES",
    time		=> 20,
    targets		=> [0, 0, 0],
    exits		=> [],
    rests		=> [],
    minims		=> [[40, 400, 160, 240, 30]],
    streams		=> [[160, 450, 2, -20, 5, 15, 25, 0], [480, 450, -2, -20, 5, 15, 25, 1]],
    demo_strokes	=> [[205, 150, 430, 210, 150, 340, 215, 150, 250, 220, 167, 233, 225, 184, 233, 230, 202, 250, 235, 184, 267, 240, 167, 267, 245, 150, 250, 250, 150, 250],
                            [300, 0, 0, 600, 0, 0]],
    text		=> [["SEMIBREVES GET FASTER WHEN", 0, 40, 3],
                            ["THEY GO THROUGH THE HEAD  ", 0, 70, 3],
                            ["OF A MINIM                ", 0, 100, 3]],
  },

  {
    name		=> "QUAVER RESTS",
    time		=> 20,
    targets		=> [0, 0, 0],
    exits		=> [],
    rests		=> [[460, 300], [480, 300], [460, 320], [480, 320]],
    minims		=> [],
    streams		=> [[160, 450, 2, -20, 5, 15, 25, 0], [480, 450, -2, -20, 5, 15, 25, 1]],
    text		=> [["RESTS SLOW DOWN NEARBY    ", 0, 40, 3],
                            ["SEMIBREVES                ", 0, 70, 3]],
  },

  {
    name		=> "VARIATIONS",
    time		=> 20,
    targets		=> [0, 0, 0],
    exits		=> [],
    rests		=> [[600, 400]],
    minims		=> [[300, 150, 480, 150, 20]],
    streams		=> [[0, 450, 15, 0, 5, 16, 25, 2]],
    gravity		=> [0, -1],
    elasticity		=> 0.9,
    slow		=> 0.1,
    text		=> [["WATCH OUT FOR LEVELS WHERE", 0, 40, 3],
                            ["SEMIBREVES BEHAVE IN AN   ", 0, 70, 3],
                            ["UNUSUAL MANNER            ", 0, 100, 3]],
  },

  {
    name		=> "TIME AND TARGETS",
    time		=> 10,
    targets		=> [1, 2, 3],
    exits		=> [],
    rests		=> [],
    minims		=> [],
    streams		=> [],
    text		=> [["THE BOTTOM OF THE SCREEN  ", 0, 40, 3],
                            ["SHOWS YOUR TARGETS AND    ", 0, 70, 3],
                            ["TIME REMAINING            ", 0, 100, 3]],
  },

  {
    name		=> "TIME AND TARGETS",
    time		=> 20,
    targets		=> [1, 2, 3],
    exits		=> [[160, 400, 0], [320, 400, 1], [480, 400, 2]],
    rests		=> [],
    minims		=> [],
    streams		=> [[640, 200, -8, 0, 2, 3, 50, 0], [640, 200, -15, 0, 4, 5, 50, 0], [640, 200, -23, 0, 6, 7, 50, 0],
                            [640, 200, -15, 0, 8, 10, 25, 1], [640, 200, -8, 0, 11, 14, 25, 2]],


    text		=> [["TO COMPLETE A LEVEL YOU   ", 0, 40, 3],
                            ["MUST GET THE TARGET NUMBER", 0, 70, 3],
                            ["OF SEMIBREVES OF EACH     ", 0, 100, 3],
                            ["COLOUR TO A SUITABLE EXIT ", 0, 130, 3]],
  },

  {
    name		=> "TIME AND TARGETS",
    time		=> 20,
    targets		=> [5, 5, 0],
    exits		=> [[160, 400, 0], [480, 400, 1]],
    rests		=> [],
    minims		=> [],
    streams		=> [[160, 110, 0, 30, 5, 15, 50, 0], [480, 110, 0, 0, 5, 15, 50, 1]],
    text		=> [["SEMIBREVES MOVING QUICKLY ", 0, 40, 3],
                            ["MAY MISS THE EXIT         ", 0, 70, 3]],
  },

  {
    name		=> "TIME AND TARGETS",
    time		=> 10,
    targets		=> [1, 2, 3],
    exits		=> [],
    rests		=> [],
    minims		=> [],
    streams		=> [],
    text		=> [["IF YOU RUN OUT OF TIME YOU", 0, 40, 3],
                            ["FAIL THE LEVEL AND MUST   ", 0, 70, 3],
                            ["TRY AGAIN FROM THE        ", 0, 100, 3],
                            ["BEGINNING                 ", 0, 130, 3]],
  },

  {
    name		=> "DRAWING MINIMS",
    time		=> 20,
    targets		=> [0, 0, 0],
    exits		=> [],
    rests		=> [],
    minims		=> [],
    streams		=> [],
    demo_strokes	=> [[130, 320, 200, 150, 320, 300, 170, 320, 400, 190, 300, 420, 210, 280, 420, 230, 260, 400, 250, 280, 380, 270, 300, 380, 290, 320, 400, 300, 320, 400],
                            [350, 0, 0, 360, 0, 0]],
    text		=> [["TO DRAW A MINIM MOVE THE  ", 0, 40, 3],
                            ["MOUSE POINTER OVER ITS    ", 0, 70, 3],
                            ["OUTLINE WHILE HOLDING A   ", 0, 100, 3],
                            ["MOUSE BUTTON              ", 0, 130, 3]],
  },
  
  {
    name		=> "DRAWING MINIMS",
    time		=> 20,
    targets		=> [0, 0, 0],
    exits		=> [],
    rests		=> [],
    minims		=> [],
    streams		=> [],
    demo_strokes	=> [[130, 320, 400, 150, 300, 380, 170, 280, 380, 190, 260, 400, 210, 280, 420, 230, 300, 420, 250, 320, 400, 270, 320, 300, 290, 320, 200, 300, 320, 200],
                            [350, 0, 0, 360, 0, 0]],
    text		=> [["YOU CAN START DRAWING FROM", 0, 40, 3],
                            ["EITHER END                ", 0, 70, 3]],
  },

  {
    name		=> "DRAWING MINIMS",
    time		=> 20,
    targets		=> [0, 0, 0],
    exits		=> [],
    rests		=> [],
    minims		=> [],
    streams		=> [],
    demo_strokes	=> [
                            [50, 100, 150, 55, 100, 200, 60, 100, 250, 65, 90, 260, 70, 80, 260, 75, 70, 250, 80, 80, 240, 85, 90, 240, 90, 100, 250, 95, 100, 250, ],
                            [100, 285, 164, 105, 250, 200, 110, 214, 235, 115, 200, 235, 120, 193, 228, 125, 193, 214, 130, 207, 214, 135, 214, 221, 140, 214, 235, 145, 214, 235, ],
                            [150, 450, 200, 155, 400, 200, 160, 350, 200, 165, 340, 190, 170, 340, 180, 175, 350, 170, 180, 360, 180, 185, 360, 190, 190, 350, 200, 195, 350, 200, ],
                            [200, 585, 235, 205, 550, 200, 210, 514, 164, 215, 514, 150, 220, 521, 143, 225, 535, 143, 230, 535, 157, 235, 528, 164, 240, 514, 164, 245, 514, 164, ],
                            [250, 100, 400, 255, 100, 350, 260, 100, 300, 265, 110, 290, 270, 120, 290, 275, 130, 300, 280, 120, 310, 285, 110, 310, 290, 100, 300, 295, 100, 300, ],
                            [300, 214, 385, 305, 250, 350, 310, 285, 314, 315, 299, 314, 320, 306, 321, 325, 306, 335, 330, 292, 335, 335, 285, 328, 340, 285, 314, 345, 285, 314, ],
                            [350, 350, 350, 355, 400, 350, 360, 450, 350, 365, 460, 360, 370, 460, 370, 375, 450, 380, 380, 440, 370, 385, 440, 360, 390, 450, 350, 395, 450, 350, ],
                            [400, 514, 314, 405, 550, 350, 410, 585, 385, 415, 585, 399, 420, 578, 406, 425, 564, 406, 430, 564, 392, 435, 571, 385, 440, 585, 385, 445, 585, 385, ],
                            [450, 0, 0, 1000, 0, 0]],
    text		=> [["YOU CAN DRAW MINIMS AT ANY", 0, 40, 3],
                            ["ANGLE                     ", 0, 70, 3]],
  },

  {
    name		=> "DRAWING MINIMS",
    time		=> 20,
    targets		=> [0, 0, 0],
    exits		=> [],
    rests		=> [],
    minims		=> [],
    streams		=> [],
    demo_strokes	=> [[50, 40, 140, 55, 40, 190, 60, 40, 240, 65, 30, 250, 70, 20, 250, 75, 10, 240, 80, 20, 230, 85, 30, 230, 90, 40, 240, 95, 40, 240, ],
                            [100, 200, 140, 105, 200, 190, 110, 200, 240, 115, 185, 250, 120, 170, 250, 125, 155, 240, 130, 170, 230, 135, 185, 230, 140, 200, 240, 145, 200, 240, ],
                            [150, 360, 115, 155, 360, 190, 160, 360, 265, 165, 350, 280, 170, 340, 280, 175, 330, 265, 180, 340, 250, 185, 350, 250, 190, 360, 265, 195, 360, 265, ],
                            [200, 520, 115, 205, 520, 190, 210, 520, 265, 215, 505, 280, 220, 490, 280, 225, 475, 265, 230, 490, 250, 235, 505, 250, 240, 520, 265, 245, 520, 265, ],
                            [250, 120, 335, 255, 120, 360, 260, 120, 385, 265, 115, 390, 270, 110, 390, 275, 105, 385, 280, 110, 380, 285, 115, 380, 290, 120, 385, 295, 120, 385, ],
                            [300, 280, 310, 305, 280, 360, 310, 280, 410, 315, 275, 415, 320, 270, 415, 325, 265, 410, 330, 270, 405, 335, 275, 405, 340, 280, 410, 345, 280, 410, ],
                            [350, 440, 310, 355, 440, 360, 360, 440, 410, 365, 426, 424, 370, 412, 424, 375, 398, 410, 380, 412, 396, 385, 426, 396, 390, 440, 410, 395, 440, 410, ],
                            [400, 600, 310, 405, 600, 360, 410, 600, 410, 415, 582, 428, 420, 564, 428, 425, 546, 410, 430, 564, 392, 435, 582, 392, 440, 600, 410, 445, 600, 410, ],
                            [450, 0, 0, 1000, 0, 0]],
    text		=> [["YOU CAN DRAW MINIMS OF    ", 0, 40, 3],
                            ["DIFFERENT SIZES           ", 0, 70, 3]],
  },

  {
    name		=> "DRAWING MINIMS",
    time		=> 20,
    targets		=> [0, 0, 0],
    exits		=> [],
    rests		=> [],
    minims		=> [],
    streams		=> [],
    demo_strokes	=> [
                            [100, 220, 160, 110, 220, 260, 120, 220, 360, 130, 240, 380, 140, 260, 380, 150, 280, 360, 160, 260, 340, 170, 240, 340, 180, 220, 360, 190, 220, 360, ],
                            [250, 430, 360, 260, 430, 260, 270, 430, 160, 280, 410, 140, 290, 390, 140, 300, 370, 160, 310, 390, 180, 320, 410, 180, 330, 430, 160, 340, 430, 160, ],
                            [400, 0, 0, 600, 0, 0]],
    text		=> [["YOU CANNOT DRAW MIRRORED  ", 0, 40, 3],
                            ["MINIMS                    ", 0, 70, 3]],
  },

  {
    name		=> "DRAWING MINIMS",
    time		=> 45,
    targets		=> [0, 0, 0],
    exits		=> [],
    rests		=> [],
    minims		=> [],
    streams		=> [],
    demo_strokes	=> [
                            [50, 80, 250, 60, 80, 280, 70, 80, 310, 80, 80, 340, 90, 80, 350, 100, 80, 360, 110, 50, 360, 120, 50, 340, 130, 80, 340, 140, 85, 340, 150, 85, 350, 160, 85, 350, ],
                            [190, 160, 250, 200, 140, 280, 210, 180, 310, 220, 160, 340, 230, 160, 350, 240, 160, 360, 250, 130, 360, 260, 130, 340, 270, 160, 340, 280, 165, 340, 290, 165, 350, 300, 165, 350, ],
                            [330, 240, 250, 340, 240, 280, 350, 240, 310, 360, 240, 340, 370, 240, 350, 380, 240, 360, 385, 210, 360, 390, 190, 360, 400, 190, 340, 405, 210, 340, 410, 240, 340, 420, 245, 340, 430, 245, 350, 440, 245, 350, ],
                            [470, 320, 250, 480, 320, 280, 490, 320, 310, 500, 320, 340, 510, 320, 350, 520, 320, 370, 530, 290, 370, 533, 290, 360, 536, 290, 340, 540, 290, 330, 550, 320, 330, 560, 325, 330, 570, 325, 350, 580, 325, 350, ],
                            [610, 400, 250, 620, 400, 280, 630, 400, 310, 640, 400, 340, 650, 400, 350, 660, 400, 360, 670, 370, 360, 680, 370, 340, 690, 400, 340, 700, 405, 340, 710, 405, 360, 720, 405, 360, ],
                            [750, 480, 250, 760, 480, 280, 770, 480, 310, 780, 480, 340, 790, 480, 350, 800, 480, 360, 810, 450, 360, 820, 450, 340, 830, 480, 340, 840, 485, 340, 850, 485, 335, 860, 485, 335, ],
                            [890, 560, 250, 900, 560, 280, 910, 560, 310, 920, 560, 340, 930, 560, 350, 940, 560, 360, 950, 530, 360, 960, 530, 340, 970, 560, 340, 980, 565, 340, 990, 565, 350, 1000, 565, 350, ],
                            [1030, 0, 0, 1250, 0, 0]],
    text		=> [["IF YOUR MINIM IS MISSHAPEN", 0, 40, 3],
                            ["THEN GUIDELINES WILL      ", 0, 70, 3],
                            ["APPEAR TO SHOW THE        ", 0, 100, 3],
                            ["EXPECTED SHAPE            ", 0, 130, 3]],
  },

  {
    name		=> "DRAWING MINIMS",
    time		=> 60,
    targets		=> [0, 0, 0],
    exits		=> [],
    rests		=> [],
    minims		=> [],
    streams		=> [],
    demo_strokes	=> [ "easy",
                            [50, 80, 250, 60, 80, 280, 70, 80, 310, 80, 80, 340, 90, 80, 350, 100, 80, 360, 110, 50, 360, 120, 50, 340, 130, 80, 340, 140, 85, 340, 150, 85, 350, 160, 85, 350, ],
                            [190, 160, 250, 200, 140, 280, 210, 180, 310, 220, 160, 340, 230, 160, 350, 240, 160, 360, 250, 130, 360, 260, 130, 340, 270, 160, 340, 280, 165, 340, 290, 165, 350, 300, 165, 350, ],
                            [330, 240, 250, 340, 240, 280, 350, 240, 310, 360, 240, 340, 370, 240, 350, 380, 240, 360, 385, 210, 360, 390, 190, 360, 400, 190, 340, 405, 210, 340, 410, 240, 340, 420, 245, 340, 430, 245, 350, 440, 245, 350, ],
                            [470, 320, 250, 480, 320, 280, 490, 320, 310, 500, 320, 340, 510, 320, 350, 520, 320, 370, 530, 290, 370, 533, 290, 360, 536, 290, 340, 540, 290, 330, 550, 320, 330, 560, 325, 330, 570, 325, 350, 580, 325, 350, ],
                            [610, 400, 250, 620, 400, 280, 630, 400, 310, 640, 400, 340, 650, 400, 350, 660, 400, 360, 670, 370, 360, 680, 370, 340, 690, 400, 340, 700, 405, 340, 710, 405, 360, 720, 405, 360, ],
                            [750, 480, 250, 760, 480, 280, 770, 480, 310, 780, 480, 340, 790, 480, 350, 800, 480, 360, 810, 450, 360, 820, 450, 340, 830, 480, 340, 840, 485, 340, 850, 485, 335, 860, 485, 335, ],
                            [890, 560, 250, 900, 560, 280, 910, 560, 310, 920, 560, 340, 930, 560, 350, 940, 560, 360, 950, 530, 360, 960, 530, 340, 970, 560, 340, 980, 565, 340, 990, 565, 350, 1000, 565, 350, ],

                            [1030, 120, 180, 1035, 120, 140, 1040, 140, 160, 1045, 160, 140, 1050, 160, 180,
                            1055, 180, 180, 1060, 200, 140, 1065, 220, 180,
                            1070, 240, 180, 1075, 240, 140, 1080, 260, 140, 1085, 280, 160, 1090, 260, 180,
                            1095, 300, 180, 1100, 300, 140, 1105, 340, 180, 1110, 340, 140,
                            1115, 400, 140, 1120, 360, 150, 1125, 400, 160, 1130, 360, 170, 1135, 400, 180,
                            1140, 460, 140, 1145, 420, 153, 1150, 460, 167, 1155, 420, 180, 1160, 460, 180,
                            1165, 520, 140, 1170, 480, 153, 1175, 520, 167, 1180, 480, 180, 1185, 480, 180,],

                            [1210, 400, 250, 1220, 360, 250, 1225, 360, 340, 1230, 360, 360, 1240, 400, 360, 1250, 400, 350, 1260, 400, 350],
                            [1290, 480, 250, 1295, 440, 260, 1300, 480, 270, 1305, 440, 280, 1310, 480, 290, 1315, 440, 300, 1320, 480, 310, 1325, 440, 320, 1330, 480, 330, 1335, 440, 340, 1340, 480, 350, 1345, 440, 360, 1350, 485, 350, 1355, 485, 350],

                            [1385, 0, 0, 1600, 0, 0]],
    text		=> [["MINIMS DO NOT NEED TO BE  ", 0, 40, 3],
                            ["DRAWN SO CAREFULLY WHEN   ", 0, 70, 3],
                            ["PLAYING ON EASY DIFFICULTY", 0, 100, 3]],
  },

  {
    name		=> "DRAWING MINIMS",
    time		=> 30,
    targets		=> [0, 0, 0],
    exits		=> [],
    rests		=> [],
    minims		=> [],
    streams		=> [],
    demo_strokes	=> [
                            [100, 320, 200, 110, 295, 265, 120, 345, 330, 130, 320, 395, 140, 320, 400, 150, 300, 420, 160, 280, 420, 170, 260, 400, 180, 280, 380, 190, 300, 380, 200, 320, 400, 210, 320, 400, ],
                            [240, 220, 200, 250, 220, 300, 260, 220, 380, 270, 160, 400, 280, 220, 420, 290, 220, 400, 300, 220, 400, ],
                            [330, 420, 200, 340, 420, 300, 350, 420, 400, 360, 360, 380, 370, 360, 420, 380, 420, 400, 390, 420, 400, ],
                            [420, 120, 200, 430, 120, 300, 440, 120, 380, 450, 60, 380, 460, 60, 420, 470, 120, 420, 480, 120, 400, 490, 120, 400, ],
                            [520, 520, 200, 530, 520, 300, 540, 520, 400, 550, 490, 380, 560, 460, 400, 570, 490, 420, 580, 520, 400, 590, 520, 400, ],
                            [620, 0, 0, 800, 0, 0]],
    text		=> [["EVEN ON HARD DIFFICULTY   ", 0, 40, 3],
                            ["YOU DO NOT NEED TO DRAW   ", 0, 70, 3],
                            ["MINIMS PERFECTLY          ", 0, 100, 3]],
  },

  {
    name		=> "DRAWING MINIMS",
    time		=> 10,
    targets		=> [0, 0, 0],
    exits		=> [],
    rests		=> [],
    minims		=> [],
    streams		=> [],
    text		=> [["THE MOST IMPORTANT POINTS ", 0, 40, 3],
                            ["OF YOUR DRAWING ARE THE   ", 0, 70, 3],
                            ["START AND END             ", 0, 100, 3]],
  },

  {
    name		=> "DRAWING MINIMS",
    time		=> 30,
    targets		=> [0, 0, 0],
    exits		=> [],
    rests		=> [],
    minims		=> [[100, 380, 540, 380, 0], [100, 400, 540, 400, 0], [100, 420, 540, 420, 0]],
    streams		=> [],
    demo_strokes	=> [
                            [100, 320, 200, 110, 320, 300, 120, 320, 400, 130, 300, 420, 140, 280, 420, 150, 260, 400, 160, 280, 380, 170, 300, 380, 180, 320, 400, 190, 320, 400, ],
                            [220, 220, 200, 230, 220, 300, 240, 220, 400, 250, 200, 420, 260, 180, 420, 270, 160, 400, 280, 180, 380, 290, 200, 380, 300, 220, 380, 310, 220, 380, ],
                            [340, 420, 200, 350, 420, 300, 360, 420, 400, 370, 400, 420, 380, 380, 420, 390, 360, 400, 400, 380, 380, 410, 400, 380, 420, 420, 420, 430, 420, 420, ],
                            [460, 220, 200, 470, 220, 300, 480, 220, 400, 490, 200, 420, 500, 180, 420, 510, 160, 400, 520, 180, 380, 530, 200, 380, 540, 200, 400, 550, 200, 400, ],
                            [580, 420, 200, 590, 420, 300, 600, 420, 400, 610, 400, 420, 620, 380, 420, 630, 360, 400, 640, 380, 380, 650, 400, 380, 660, 450, 400, 670, 450, 400, ],
                            [700, 0, 0, 800, 0, 0]],
    text		=> [["AIM TO START OR FINISH    ", 0, 40, 3],
                            ["LEVEL WITH THE MIDDLE OF  ", 0, 70, 3],
                            ["THE HEAD OF THE MINIM AND ", 0, 100, 3],
                            ["ON THE STEM OF THE MINIM  ", 0, 130, 3]],
  },

  {
    name		=> "DRAWING MINIMS",
    time		=> 20,
    targets		=> [0, 0, 0],
    exits		=> [],
    rests		=> [],
    minims		=> [[220, 370, 220, 430, 0], [220, 430, 160, 430, 0], [160, 430, 160, 370, 0], [160, 370, 220, 370, 0],
                            [320, 370, 320, 430, 0], [320, 430, 260, 430, 0], [260, 430, 260, 370, 0], [260, 370, 320, 370, 0],
                            [420, 370, 420, 430, 0], [420, 430, 360, 430, 0], [360, 430, 360, 370, 0], [360, 370, 420, 370, 0]],
    streams		=> [],
    demo_strokes	=> [
                            [100, 220, 200, 110, 220, 300, 120, 220, 400, 130, 190, 430, 140, 160, 400, 150, 190, 370, 160, 220, 400, 170, 220, 400, ],
                            [200, 320, 200, 210, 320, 300, 220, 320, 400, 230, 290, 435, 240, 260, 400, 250, 290, 365, 260, 320, 400, 270, 320, 400, ],
                            [300, 420, 200, 310, 420, 300, 320, 420, 400, 330, 385, 430, 340, 350, 400, 350, 385, 370, 360, 420, 400, 370, 420, 400, ],
                            [400, 0, 0, 600, 0, 0]],
    text		=> [["AIM TO DRAW THE HEAD OF   ", 0, 40, 3],
                            ["THE MINIM WIDER THAN IT IS", 0, 70, 3],
                            ["TALL                      ", 0, 100, 3]],
  },

  {
    name		=> "CONTROLS",
    time		=> 30,
    targets		=> [0, 0, 0],
    exits		=> [],
    rests		=> [],
    minims		=> [],
    streams		=> [],
    text		=> [["AT ANY TIME               ", 0, 50, 3],
                            ["F      TOGGLE FULLSCREEN  ", 0, 90, 3],
                            ["D      TOGGLE DIFFICULTY  ", 0, 120, 3],
                            ["DURING INTRO              ", 0, 170, 3],
                            ["ESC    QUIT GAME          ", 0, 210, 3],
                            ["MOUSE  CLICK TO START GAME", 0, 240, 3],
                            ["CREATE CUSTOM GAME LEVELS ", 0, 290, 3],
                            ["BY EDITING THE LEVELS FILE", 0, 320, 3],
                            ["IN THE GAME DIRECTORY     ", 0, 350, 3]],
  },

  {
    name		=> "CONTROLS",
    time		=> 30,
    targets		=> [0, 0, 0],
    exits		=> [],
    rests		=> [],
    minims		=> [],
    streams		=> [],
    text		=> [["AT LEVEL START            ", 0, 50, 3],
                            ["ESC    RETURN TO INTRO    ", 0, 90, 3],
                            ["R      RECORD LEVEL       ", 0, 120, 3],
                            ["P      CHOOSE RECORDING   ", 0, 150, 3],
                            ["       FOR PLAYBACK       ", 0, 180, 3],
                            ["MOUSE  CLICK TO START PLAY", 0, 210, 3],
                            ["EVERY LEVEL COMES WITH A  ", 0, 260, 3],
                            ["RECORDED SOLUTION         ", 0, 290, 3],
                            ["RECORDINGS ARE STORED IN  ", 0, 340, 3],
                            ["THE RECORD FILE IN THE    ", 0, 370, 3],
                            ["GAME DIRECTORY            ", 0, 400, 3]],
  },

  {
    name		=> "CONTROLS",
    time		=> 30,
    targets		=> [0, 0, 0],
    exits		=> [],
    rests		=> [],
    minims		=> [],
    streams		=> [],
    text		=> [["DURING PLAY               ", 0, 50, 3],
                            ["ESC    RESTART LEVEL      ", 0, 90, 3],
                            ["S      SKIP TO NEXT LEVEL ", 0, 120, 3],
                            ["MOUSE  DRAG TO DRAW MINIMS", 0, 150, 3],
                            ["FOR THE BEST EXPERIENCE   ", 0, 200, 3],
                            ["USE A GRAPHICS TABLET     ", 0, 230, 3],
                            ["INSTEAD OF A MOUSE        ", 0, 260, 3]],
  },

  {
    name		=> "CREDITS",
    time		=> 30,
    targets		=> [0, 0, 0],
    exits		=> [],
    rests		=> [],
    minims		=> [],
    streams		=> [],
    text		=> [["MINIM MADNESS WAS WRITTEN ", 0, 50, 3],
                            ["BY MARTIN MARIUSZ LESTER  ", 0, 80, 3],
                            ["IN APRIL 2008 FOR THE 11TH", 0, 110, 3],
                            ["LUDUMDARE 48 HOUR SOLO    ", 0, 140, 3],
                            ["GAME DEVELOPMENT          ", 0, 170, 3],
                            ["COMPETITION               ", 0, 200, 3],
                            ["THE COMPETITION THEME WAS ", 0, 250, 3],
                            ["MINIMALISM                ", 0, 280, 3],
                            ["VERSION 1.0 WAS RELEASED  ", 0, 330, 3],
                            ["IN MAY WITH VARIOUS       ", 0, 360, 3],
                            ["IMPROVEMENTS              ", 0, 390, 3]],
  },

  {
    name		=> "CREDITS",
    time		=> 30,
    targets		=> [0, 0, 0],
    exits		=> [],
    rests		=> [],
    minims		=> [],
    streams		=> [],
    text		=> [["DOWNLOAD THE LATEST       ", 0, 50, 3],
                            ["VERSION FROM THE BEST     ", 0, 80, 3],
                            ["NOTION SOFTWARE WEBSITE   ", 0, 110, 3],
                            ["WWW.BESTNOTION.NET        ", 0, 140, 3],
                            ["VISIT THE LUDUMDARE       ", 0, 190, 3],
                            ["COMPETITION WEBSITE       ", 0, 220, 3],
                            ["WWW.LUDUMDARE.COM         ", 0, 250, 3],
                            ["YOU MAY REDISTRIBUTE THIS ", 0, 300, 3],
                            ["GAME FREELY               ", 0, 330, 3],
                            ["SOURCE CODE IS UNDER THE  ", 0, 380, 3],
                            ["MIT X CONSORTIUM LICENCE  ", 0, 410, 3]],
  },

);


# Game level data.
@game_levels = (

  {
    name		=> "SIDESTEP LEFT",
    time		=> 100,
    targets		=> [5, 0, 0],
    exits		=> [[160, 360, 0]],
    rests		=> [],
    minims		=> [],
    streams		=> [[480, 30, 0, 0, 0, 100, 100, 0]],
    demo_strokes	=> [
                            [62, 536, 252, 67, 500, 316, 72, 464, 380, 77, 446, 386, 82, 434, 379, 87, 429, 361, 92, 447, 356, 97, 459, 362, 102, 464, 380, 107, 464, 380, ],
                            [179, 428, 407, 184, 367, 411, 189, 305, 416, 194, 292, 405, 199, 292, 393, 204, 303, 380, 209, 316, 391, 214, 316, 403, 219, 305, 416, 224, 305, 416, ],
                            [279, 134, 407, 284, 134, 350, 289, 135, 292, 294, 144, 283, 299, 154, 283, 304, 163, 292, 309, 154, 301, 314, 144, 301, 319, 135, 292, 324, 135, 292, ],
                            [540, 291, 417, 545, 236, 415, 550, 180, 412, 555, 174, 405, 560, 174, 399, 565, 181, 392, 570, 187, 399, 575, 187, 406, 580, 180, 412, 585, 180, 412, ],
                            [668, 339, 439, 673, 295, 436, 678, 250, 433, 683, 245, 427, 688, 245, 421, 693, 251, 415, 698, 257, 422, 703, 256, 428, 708, 250, 433, 713, 250, 433, ],
                            [813, 175, 416, 818, 140, 415, 823, 105, 413, 828, 98, 406, 833, 99, 399, 838, 106, 392, 843, 113, 400, 848, 112, 406, 853, 105, 413, 858, 105, 413, ],
                           ],
    text		=> [],
  },

  {
    name		=> "SIDESTEP RIGHT",
    time		=> 100,
    targets		=> [0, 5, 0],
    exits		=> [[480, 360, 1]],
    rests		=> [],
    minims		=> [],
    streams		=> [[160, 30, 0, 0, 0, 100, 100, 1]],
    demo_strokes	=> [
                            [52, 211, 167, 57, 179, 189, 62, 146, 211, 67, 142, 210, 72, 140, 207, 77, 141, 202, 82, 145, 203, 87, 147, 206, 92, 146, 211, 97, 146, 211, ],
                            [217, 138, 248, 222, 128, 195, 227, 118, 142, 232, 132, 122, 237, 148, 119, 242, 168, 133, 247, 154, 152, 252, 138, 155, 257, 118, 142, 262, 118, 142, ],
                            [467, 478, 377, 472, 392, 341, 477, 306, 305, 482, 304, 299, 487, 306, 294, 492, 312, 291, 498, 315, 297, 502, 313, 302, 507, 306, 305, 512, 306, 305, ],
                            [669, 528, 301, 674, 530, 363, 679, 532, 425, 684, 526, 431, 689, 520, 431, 694, 514, 426, 699, 520, 419, 704, 526, 419, 709, 532, 425, 714, 532, 425, ],    
                           ],

    text		=> [],
  },

  {
    name		=> "EVAPORATION",
    time		=> 60,
    targets		=> [0, 0, 10],
    exits		=> [[320, 40, 2]],
    rests		=> [],
    minims		=> [],
    streams		=> [[320, 430, 0, -5, 0, 100, 25, 2]],
    demo_strokes	=> [
                            [71, 347, 273, 76, 346, 358, 81, 345, 444, 86, 327, 462, 91, 309, 461, 96, 291, 443, 101, 309, 426, 106, 327, 426, 111, 345, 444, 116, 345, 444, ],
                            [150, 342, 247, 155, 343, 331, 160, 345, 416, 165, 322, 440, 170, 298, 441, 175, 274, 417, 180, 297, 393, 185, 321, 393, 190, 345, 416, 195, 345, 416, ],
                            [237, 338, 216, 242, 340, 301, 247, 343, 386, 252, 322, 409, 257, 300, 409, 262, 277, 388, 267, 298, 365, 272, 320, 365, 277, 343, 386, 282, 343, 386, ],
                            [314, 337, 209, 319, 337, 278, 324, 336, 348, 329, 314, 369, 334, 293, 369, 339, 272, 348, 344, 293, 326, 349, 315, 326, 354, 336, 348, 359, 336, 348, ],
                            [422, 340, 173, 427, 340, 249, 432, 341, 326, 437, 319, 349, 442, 296, 349, 447, 274, 326, 452, 296, 304, 457, 318, 304, 462, 341, 326, 467, 341, 326, ],
                           ],    
    text		=> [],
  },

  {
    name		=> "STEPPING STONE",
    time		=> 60,
    targets		=> [5, 0, 0],
    exits		=> [[600, 240, 0]],
    rests		=> [],
    minims		=> [],
    streams		=> [[0, 50, 5, 0, 0, 60, 25, 0]],
    demo_strokes	=> [
                            [249, 604, 250, 254, 324, 250, 259, 43, 250, 264, 34, 241, 269, 34, 231, 274, 43, 222, 279, 52, 231, 284, 52, 241, 289, 43, 250, 294, 43, 250, ],
                           ],
    text		=> [],
  },


  {
    name		=> "SANDSTORM",
    time		=> 100,
    targets		=> [0, 0, 5],
    exits		=> [[160, 360, 2]],
    rests		=> [],
    minims		=> [],
    streams		=> [[480, 30, 0, 0, 0, 100, 100, 2],
                            [0, 100, 29, 0, 10, 100, 28, 0],
                            [0, 200, 30, -10, 10, 100, 39, 0],
                            [0, 300, 31, -20, 10, 100, 25, 0],
                            [0, 150, 10, 0, 10, 100, 29, 0],
                            [0, 250, 17, -9, 10, 100, 23, 0],
                            [0, 350, 13, -22, 10, 100, 21, 0]],
    demo_strokes	=> [
                            [71, 533, 278, 76, 486, 352, 81, 438, 425, 86, 425, 428, 91, 416, 423, 96, 413, 409, 101, 427, 406, 106, 435, 411, 111, 438, 425, 116, 438, 425, ],
                            [230, 415, 432, 235, 347, 431, 240, 279, 429, 245, 268, 418, 250, 268, 407, 255, 280, 396, 260, 291, 407, 265, 290, 418, 270, 279, 429, 275, 279, 429, ],
                            [570, 133, 421, 575, 130, 357, 580, 126, 293, 585, 134, 284, 590, 142, 284, 595, 151, 292, 600, 143, 301, 605, 135, 301, 610, 126, 293, 615, 126, 293, ],
                            [1026, 98, 394, 1031, 130, 354, 1036, 162, 314, 1041, 172, 313, 1046, 177, 317, 1051, 178, 326, 1056, 169, 327, 1061, 163, 323, 1066, 162, 314, 1071, 162, 314, ],
                           ],
    text		=> [],
  },

  {
    name		=> "IN A NICK OF TIME",
    time		=> 20,
    targets		=> [20, 0, 0],
    exits		=> [[320, 430, 0], [320, 450, 0]],
    rests		=> [],
    minims		=> [],
    streams		=> [[320, 30, 0, 0, 0, 20, 25, 0]],
    demo_strokes	=> [
                            [42, 353, 50, 47, 360, 135, 52, 368, 220, 57, 346, 246, 62, 323, 248, 67, 297, 227, 72, 318, 201, 77, 342, 199, 82, 368, 220, 87, 368, 220, ],
                            [127, 348, 33, 132, 357, 108, 137, 367, 184, 142, 346, 210, 147, 323, 213, 152, 296, 193, 157, 317, 166, 162, 340, 163, 167, 367, 184, 172, 367, 184, ],
                            [368, 340, 29, 373, 340, 88, 378, 341, 148, 383, 324, 165, 388, 307, 166, 393, 289, 148, 398, 306, 131, 403, 324, 131, 408, 341, 148, 413, 341, 148, ],
                           ],
    text		=> [],
  },


  {
    name		=> "CROSSOVER",
    time		=> 100,
    targets		=> [10, 10, 0],
    exits		=> [[160, 420, 0], [480, 420, 1]],
    rests		=> [],
    minims		=> [],
    streams		=> [[160, 30, 0, 0, 0, 100, 25, 1], [480, 30, 0, 0, 0, 100, 25, 0]],
    demo_strokes	=> [
                            [100, 195, 205, 105, 167, 225, 110, 138, 246, 115, 129, 244, 120, 125, 239, 125, 127, 230, 130, 136, 231, 135, 140, 237, 140, 138, 246, 145, 138, 246, ],
                            [216, 117, 304, 221, 112, 236, 226, 107, 167, 231, 121, 151, 236, 137, 150, 241, 153, 164, 246, 139, 181, 251, 124, 182, 256, 107, 167, 261, 107, 167, ],
                            [503, 169, 33, 508, 172, 72, 513, 176, 111, 518, 165, 124, 523, 153, 125, 528, 140, 114, 533, 151, 101, 538, 163, 100, 543, 176, 111, 548, 176, 111, ],
                            [668, 160, 30, 673, 166, 69, 678, 172, 108, 683, 162, 122, 688, 150, 124, 693, 136, 114, 698, 146, 100, 703, 158, 98, 708, 172, 108, 713, 172, 108, ],
                            [886, 531, 166, 891, 486, 210, 896, 440, 254, 901, 427, 253, 906, 421, 247, 911, 421, 234, 916, 434, 234, 921, 441, 240, 926, 440, 254, 931, 440, 254, ],
                            [1014, 223, 455, 1019, 171, 459, 1024, 119, 463, 1029, 110, 455, 1034, 110, 447, 1039, 117, 438, 1044, 126, 446, 1049, 127, 454, 1054, 119, 463, 1059, 119, 463, ],
                           ],
    text		=> [],
  },


  {
    name		=> "DOWN UNDER",
    time		=> 50,
    targets		=> [0, 10, 0],
    exits		=> [[480, 440, 1]],
    rests		=> [],
    minims		=> [],
    streams		=> [[160, 440, 5, 0, 0, 50, 25, 1]],
    demo_strokes	=> [
                            [45, 183, 228, 50, 289, 228, 55, 396, 229, 60, 405, 238, 65, 405, 246, 70, 396, 255, 75, 387, 246, 80, 387, 238, 85, 396, 229, 90, 396, 229, ],
                            [139, 131, 456, 144, 161, 402, 149, 192, 348, 154, 215, 341, 159, 230, 350, 164, 237, 373, 169, 214, 380, 174, 199, 372, 179, 192, 348, 184, 192, 348, ],
                            [240, 134, 449, 245, 163, 396, 250, 193, 343, 255, 214, 337, 260, 228, 345, 265, 234, 366, 270, 212, 372, 275, 199, 364, 280, 193, 343, 285, 193, 343, ],
                            [383, 204, 342, 388, 219, 322, 393, 234, 302, 398, 244, 301, 403, 249, 305, 408, 251, 314, 413, 241, 316, 418, 235, 312, 423, 234, 302, 428, 234, 302, ],
                            [507, 226, 312, 512, 238, 290, 517, 250, 268, 522, 260, 265, 527, 267, 269, 532, 270, 279, 537, 260, 282, 542, 253, 278, 547, 250, 268, 552, 250, 268, ],
                           ],
    text		=> [],
    gravity		=> [0, -1],
  },


  {
    name		=> "JUST IN TIME",
    time		=> 20,
    targets		=> [0, 20, 0],
    exits		=> [[320, 430, 1]],
    rests		=> [],
    minims		=> [],
    streams		=> [[320, 30, 0, 0, 0, 20, 25, 1]],
    demo_strokes	=> [
                            [186, 363, 463, 191, 320, 463, 196, 277, 464, 201, 272, 459, 206, 272, 454, 211, 277, 448, 216, 282, 454, 221, 282, 459, 226, 277, 464, 231, 277, 464, ],
                            [312, 346, 18, 317, 349, 74, 322, 353, 131, 327, 337, 149, 332, 320, 150, 337, 301, 134, 342, 318, 116, 347, 335, 115, 352, 353, 131, 357, 353, 131, ],
                            [373, 343, 18, 378, 346, 74, 383, 350, 130, 388, 334, 148, 393, 317, 149, 398, 298, 133, 403, 314, 115, 408, 332, 114, 413, 350, 130, 418, 350, 130, ],
                            [427, 338, 21, 432, 346, 78, 437, 354, 135, 442, 340, 154, 447, 323, 156, 452, 304, 142, 457, 318, 123, 462, 335, 120, 467, 354, 135, 472, 354, 135, ],
                           ],
    text		=> [],
  },


  {
    name		=> "CROCODILES",
    time		=> 60,
    targets		=> [5, 0, 0],
    exits		=> [[600, 240, 0]],
    rests		=> [],
    minims		=> [],
    streams		=> [[0, 50, 5, 0, 0, 60, 25, 0], [100, 450, 1, -25, 0, 60, 25, 1], [300, 450, 1, -25, 0, 60, 25, 1], [500, 450, 1, -25, 0, 60, 25, 1]],
    demo_strokes	=> [
                            [65, 92, 123, 70, 65, 125, 75, 38, 128, 80, 33, 123, 85, 32, 118, 90, 37, 113, 95, 42, 117, 100, 43, 122, 105, 38, 128, 110, 38, 128, ],
                            [220, 11, 24, 225, 36, 42, 230, 61, 60, 235, 62, 72, 240, 57, 79, 245, 45, 81, 250, 43, 69, 255, 48, 62, 260, 61, 60, 265, 61, 60, ],
                            [393, 285, 161, 398, 217, 154, 403, 150, 147, 408, 143, 139, 413, 144, 131, 418, 152, 124, 423, 159, 133, 428, 158, 140, 433, 150, 147, 438, 150, 147, ],
                            [639, 496, 207, 644, 428, 206, 649, 359, 205, 654, 352, 197, 659, 352, 190, 664, 359, 182, 669, 367, 190, 674, 367, 198, 679, 359, 205, 684, 359, 205, ],
                            [1064, 371, 208, 1069, 349, 207, 1074, 327, 206, 1079, 321, 200, 1084, 322, 194, 1089, 328, 188, 1094, 333, 195, 1099, 333, 200, 1104, 327, 206, 1109, 327, 206, ],
                            [1170, 627, 114, 1175, 632, 198, 1180, 638, 282, 1185, 632, 289, 1190, 625, 289, 1195, 617, 283, 1200, 624, 276, 1205, 631, 275, 1210, 638, 282, 1215, 638, 282, ],
                           ],
    text		=> [],
  },

  {
    name		=> "EXCHANGE",
    time		=> 100,
    targets		=> [10, 10, 0],
    exits		=> [[160, 420, 0], [480, 420, 1]],
    rests		=> [],
    minims		=> [],
    streams		=> [[160, 440, 0, -10, 0, 100, 25, 1], [480, 440, 0, -10, 0, 100, 25, 0]],
    demo_strokes	=> [
                            [106, 151, 469, 111, 152, 452, 116, 153, 435, 121, 158, 431, 126, 163, 431, 131, 167, 436, 136, 162, 440, 141, 158, 440, 146, 153, 435, 151, 153, 435, ],
                            [164, 154, 472, 169, 153, 453, 174, 151, 434, 179, 155, 430, 184, 159, 429, 189, 163, 433, 194, 160, 438, 199, 155, 438, 204, 151, 434, 209, 151, 434, ],
                            [293, 153, 468, 298, 153, 452, 303, 153, 435, 308, 158, 430, 313, 163, 430, 318, 167, 435, 323, 163, 440, 328, 158, 440, 333, 153, 435, 338, 153, 435, ],
                            [436, 154, 471, 441, 154, 452, 446, 154, 433, 451, 158, 429, 456, 162, 429, 461, 166, 433, 466, 162, 437, 471, 158, 437, 476, 154, 433, 481, 154, 433, ],
                            [516, 155, 468, 521, 155, 451, 526, 154, 433, 531, 158, 429, 536, 162, 429, 541, 166, 433, 546, 162, 437, 551, 158, 437, 556, 154, 433, 561, 154, 433, ],
                            [728, 97, 398, 733, 146, 339, 738, 195, 280, 743, 214, 279, 748, 225, 287, 753, 227, 307, 758, 207, 309, 763, 197, 300, 768, 195, 280, 773, 195, 280, ],
                            [898, 469, 428, 903, 332, 398, 908, 194, 368, 913, 187, 357, 918, 189, 348, 923, 200, 341, 928, 208, 352, 933, 206, 361, 938, 194, 368, 943, 194, 368, ],
                            [1216, 498, 256, 1221, 508, 339, 1226, 518, 423, 1231, 502, 443, 1236, 485, 445, 1241, 465, 429, 1246, 480, 409, 1251, 498, 407, 1256, 518, 423, 1261, 518, 423, ],
                            [1332, 392, 204, 1337, 449, 262, 1342, 507, 321, 1347, 507, 344, 1352, 495, 355, 1357, 472, 355, 1362, 472, 332, 1367, 484, 321, 1372, 507, 321, 1377, 507, 321, ],
                            [1436, 367, 206, 1441, 439, 275, 1446, 512, 344, 1451, 512, 360, 1456, 504, 369, 1461, 488, 369, 1466, 487, 353, 1471, 495, 344, 1476, 512, 344, 1481, 512, 344, ],
                            [1714, 147, 431, 1719, 104, 414, 1724, 62, 396, 1729, 57, 386, 1734, 60, 378, 1739, 71, 373, 1744, 76, 384, 1749, 73, 392, 1754, 62, 396, 1759, 62, 396, ],
                           ],
    text		=> [],
  },

  {
    name		=> "CROSS COUNTRY",
    time		=> 100,
    targets		=> [5, 0, 0],
    exits		=> [[105, 430, 0]],
    rests		=> [],
    minims		=> [[213, 450, 213, 170, 0], [213, 170, 213, 450, 0],
                            [426, 30, 426, 310, 0], [426, 310, 426, 30, 0],
                            [0, 30, 426, 30, 0], [426, 30, 0, 30, 0],
                            [213, 450, 640, 450, 0], [640, 450, 213, 450, 0]],
    streams		=> [[533, 50, 0, 0, 0, 100, 25, 0]],
    demo_strokes	=> [
                            [9, 571, 341, 14, 529, 394, 19, 487, 446, 24, 472, 448, 29, 465, 442, 34, 463, 428, 39, 477, 426, 44, 485, 432, 49, 487, 446, 54, 487, 446, ],
                            [55, 568, 350, 60, 534, 397, 65, 499, 445, 70, 486, 447, 75, 479, 441, 80, 477, 428, 85, 490, 426, 90, 497, 432, 95, 499, 445, 100, 499, 445, ],
                            [139, 573, 99, 144, 581, 212, 149, 589, 324, 154, 563, 354, 159, 536, 356, 164, 506, 330, 169, 532, 301, 174, 559, 299, 179, 589, 324, 184, 589, 324, ],
                            [332, 423, 312, 337, 394, 343, 342, 364, 375, 347, 353, 375, 352, 348, 370, 357, 348, 359, 362, 358, 359, 367, 364, 364, 372, 364, 375, 377, 364, 375, ],
                            [480, 387, 118, 485, 392, 198, 490, 397, 279, 495, 384, 293, 500, 371, 294, 505, 357, 281, 510, 369, 267, 515, 383, 266, 520, 397, 279, 525, 397, 279, ],
                            [624, 237, 59, 629, 304, 93, 634, 371, 128, 639, 379, 151, 644, 371, 166, 649, 348, 174, 654, 340, 151, 659, 348, 135, 664, 371, 128, 669, 371, 128, ],
                            [786, 107, 470, 791, 73, 455, 796, 39, 440, 801, 36, 431, 806, 39, 425, 811, 47, 422, 816, 51, 431, 821, 48, 437, 826, 39, 440, 831, 39, 440, ],
                           ],
    text		=> [],
  },

  {
    name		=> "CONGRATULATIONS",
    time		=> 60,
    targets		=> [0, 0, 0],
    exits		=> [],
    rests		=> [],
    minims		=> [],
    streams		=> [[320, 450, 2, -25, 0, 60, 26, 2], [320, 450, 4, -25, 0, 60, 24, 2], [320, 450, -2, -25, 0, 60, 27, 2], [320, 450, -4, -25, 0, 60, 23, 2],
                            [0, 240, 3, -15, 0, 60, 26, 0], [0, 240, 5, -15, 0, 60, 24, 0], [0, 240, 3, -10, 0, 60, 27, 0], [0, 240, 5, -10, 0, 60, 23, 0],
                            [639, 240, -3, -15, 0, 60, 26, 1], [639, 240, -5, -15, 0, 60, 24, 1], [639, 240, -3, -10, 0, 60, 27, 1], [639, 240, -5, -10, 0, 60, 23, 1]],
    text		=> [],
  },



) if (!$custom);

# slalom
# kitchen sink
# bunker
# quicksand
# velvet glove



# Dump levels to text file.
# This is useful for testing the levels file parser or for producing a levels file for the built-in levels.
while (0) {
  my $levels = new IO::File("leveldump", ">");
  my @colours = ("red", "green", "blue");
  foreach my $level (@demo_levels) {

    $levels->print("level " . $level->{name} . "\n");
    $levels->print("time " . $level->{time} . "\n");
    $levels->print("target red " . $level->{targets}->[0] . "\n");
    $levels->print("target green " . $level->{targets}->[1] . "\n");
    $levels->print("target blue " . $level->{targets}->[2] . "\n");
    $levels->print("gravity " . $level->{gravity}->[0] . " " . $level->{gravity}->[1] . "\n") if defined($level->{gravity});
    $levels->print("elasticity " . $level->{elasticity} . "\n") if defined($level->{elasticity});
    $levels->print("boost " . $level->{boost} . "\n") if defined($level->{boost});
    $levels->print("slow " . $level->{slow} . "\n") if defined($level->{slow});

    foreach my $exit (@{$level->{exits}}) {
      $levels->print("exit " . $colours[$exit->[2]] . " " . $exit->[0] . " " . $exit->[1] . "\n");
    }

    foreach my $rest (@{$level->{rests}}) {
      $levels->print("rest " . join(" ", @{$rest}) . "\n");
    }

    foreach my $minim (@{$level->{minims}}) {
      $levels->print("minim " . join(" ", @{$minim}) . "\n");
    }

    foreach my $stream (@{$level->{streams}}) {
      $levels->print("stream " . $colours[$stream->[7]] . " " . join(" ", @{$stream}[0 .. 6]) . "\n");
    }

    foreach my $label (@{$level->{text}}) {
      $levels->print("text " . join(" ", @{$label}[(1, 2, 3, 0)]) . "\n");
    }

    foreach my $stroke (@{$level->{demo_strokes}}) {
      if (ref($stroke)) {
        $levels->print("stroke " . join(" ", @{$stroke}) . "\n");
      }
      else {
        $levels->print("stroke $stroke\n");
      }
    }
    
    $levels->print("\n");
  }
  $levels->close();
}

# Start in demo mode, on first demo.
my $demo = 1;
my $stage = 0;

# Turn off easy drawing by default.
my $easy = 0;
# Toggle difficulty and play a sound.
sub toggle_difficulty {
  if ($easy) {
    $easy = 0;
    &$play("hard");
  }
  else {
    $easy = 1;
    &$play("easy");
  }
}

OUTER_LOOP: while (1) {

  # Prepare for main game loop.

  # Note that the note surface needs to be drawn.
  my $redraw_note = 1;

  # Blank the main surfaces at the start of a level.
  $bg_surf->fill($app_rect, $black);
  $note_surf->fill($app_rect, $black);
  $draw_surf->fill($app_rect, $black);
  $app->fill($app_rect, $black);


  # Get level data and unpack it.
  my $level;
  if ($demo) {
    $level = dclone($demo_levels[$stage]);
  }
  else {
    $level = dclone($game_levels[$stage]);
  }

  # Fake mouse strokes for demo levels and recordings.
  my @demo_strokes = ();
  @demo_strokes = @{$level->{demo_strokes}} if (defined($level->{demo_strokes}));
  my $demo_easy = 0;

  # The time limit and number of seconds elapsed.
  my $time_limit = $level->{time} || 100;
  my $seconds = -1;

  # Variable physical properties.
  my $elasticity = defined($level->{elasticity}) ? $level->{elasticity} : 0.6;
  my $boost = defined($level->{boost}) ? $level->{boost} : 1.1;
  my $slow = defined($level->{slow}) ? $level->{slow} : 0.9;
  my @gravity = defined($level->{gravity}) ? @{$level->{gravity}} : (0, 1);

  # Target numbers of semibreves of each colour.
  my @targets = @{$level->{targets}};

  # Initialise any initial minims properly.
  my @minims = ();
  foreach my $minim (@{$level->{minims}}) {
    my ($x1, $y1, $x2, $y2, $r) = @{$minim};
    my $length = distance($x1, $y1, $x2, $y2);
    my $dx = $x2 - $x1;
    my $dy = $y2 - $y1;
    my $angle = atan2($dy, $dx);
    my $scale = $r/$length;
    my $x3 = int($x2 - ($dy*$scale));
    my $y3 = int($y2 + ($dx*$scale));
    push (@minims, [$x1, $y1, $x2, $y2, $x3, $y3, $r, $length, $angle]);
  }

  # Streams of semibreves for this level.
  my @streams = @{$level->{streams}};
  my @active_streams = ();

  # Draw quaver rests on background.
  my @rests = @{$level->{rests}};
  foreach my $sniff (@rests) {
    $rest_targ->x($sniff->[0] - 10);
    $rest_targ->y($sniff->[1] - 10);
    $rest->blit($rest_rect, $bg_surf, $rest_targ);
  }  

  # Draw exits on background.
  my @exits = @{$level->{exits}};
  foreach my $exit (@exits) {
    $box_targ->x($exit->[0] - 10);
    $box_targ->y($exit->[1] - 10);
    $boxes[$exit->[2]]->blit($box_rect, $bg_surf, $box_targ);
  }

  # Print any text labels on the background.
  my @texts = @{$level->{text}};
  foreach my $text (@texts) {
    my ($label, $x, $y, $col) = @{$text};
    font_print($fonts[$col], $bg_surf, $label, $x, $y);
  }

  # Print level name, initial targets and time.
  my $name = $level->{name};
  font_print($fonts[3], $bg_surf, $name, 0, 0);
  font_print($fonts[3], $bg_surf, "TARGETS", 5, 460);
  foreach my $t (0..2) {
    font_print($fonts[$t], $bg_surf, sprintf("%03d", $targets[$t]), 185 + (85*$t), 460);  
  }
  font_print($fonts[3], $bg_surf, "TIME", 455, 460);
  font_print($fonts[3], $bg_surf, sprintf("%3d", $time_limit), 565, 460);

  # Keep going until the level ends.
  my $end_game = 0;
  
  # The level can end due to skipping, running out of time, starting a new game, success or aborting.
  my ($level_skip, $level_timeover, $level_newgame, $level_success, $level_restart) = (0,0,0,0,0);

  # For line drawing we need to know which buttons are pressed and when.
  my $mouse_buttons = 0;
  my @lines = ();
  my @points = ();

  # No semibreves to start with.
  my @semibreves = ();

  # Allow for recording and playback of demos/solutions.
  my $record = 0;
  my $record_easy = 0;
  my $playback = 0;
  my $recording;

  # Invite the player to click to start game levels.
  if (!$demo) {
    # Need to draw notes.
    $bg_surf->blit($app_rect, $app, $app_rect);
    if ($redraw_note) {
      $note_surf->fill($app_rect, $black);
      foreach my $minim (@minims) {
        draw_minim($note_surf, $cyan, @{$minim});
      }
      $redraw_note = 0;
    }
    $note_surf->blit($app_rect, $app, $app_rect);
    $draw_surf->blit($app_rect, $app, $app_rect);

    # Display a notice.
    notice($app, $fonts[3], "CLICK TO START", $white, $black);
    $app->sync();

    # Allow player to choose recording played back.
    my @recordings = ();
    my $solution_avail = 0;
    my $records_avail = undef;
    my $recording_chosen = 0;
    my $no_record_file = 0;
    my $corrupt_record_file = 0;

    # Wait for click.
    while ($event->wait()) {
      my $type = $event->type();
      if ($type == SDL_QUIT) {
        # Quit immediately.
        exit(0);
      }
      elsif ($type == SDL_MOUSEBUTTONUP) {
        # Start the game.
        &$play("click") if (!$demo);
        last;
      }
      elsif ($type == SDL_KEYUP && $event->key_name() eq "escape") {
        # Back to demo.
        &$play("click") if (!$demo);
        $demo = 1;
        $stage = 0;
        next OUTER_LOOP;
      }
      elsif ($type == SDL_KEYUP && $event->key_name() eq "f") {
        # Toggle fullscreen.
        $app->fullscreen();
      }
      elsif ($type == SDL_KEYUP && $event->key_name() eq "d") {
        # Toggle difficulty.
        toggle_diffculty();
      }      
      elsif ($type == SDL_KEYUP && $event->key_name() eq "r") {
        # Start recording level.
        &$play("click") if (!$demo);
        $recording = new IO::File("record", ">>");
        if (defined($recording)) {
          $record = 1;
          $playback = 0;
          $recording->print("\n# Recording level $stage ($name) at " . localtime() . ".\n\n");
          $recording->print("level $name\n");
          last;
        }
        else {
          notice($app, $fonts[3], "CANNOT OPEN RECORD FILE", $white, $black);
          $app->sync();
        }
      }      
      elsif ($type == SDL_KEYUP && $event->key_name() eq "p") {
        # Playback.
        &$play("click") if (!$demo);
        
        # If this is the first time P has been pressed, look for recordings.
        if (!defined($records_avail)) {
        
          $records_avail = 0;

          # First look for a solution in the level definition.
          if (scalar @demo_strokes > 0) {
            $solution_avail = 1;
            push (@recordings, [@demo_strokes]);
          }
          else {
            push (@recordings, []);
          }
          
          # Try to find recordings in the recording file.
          {
            $recording = new IO::File("record", "<");
            if (!defined($recording)) {
              $no_record_file = 1;
              last;
            }

            my $reading = 0;
            my @strokes = ();
            my $linenum = 0;
            while (!$recording->eof()) {
              $linenum++;
              my $line = $recording->getline();

              # Strip comments.
              if ($line =~ m/^(.*?)#/ ) {
                $line = $1;
              }
              
              # Ignore blank lines.
              next if ($line =~ m/^\s*$/ );

              # Strip leading and trailing space.
              if ($line =~ m/^(.*?)\s+$/ ) {
                $line = $1;
              }
              if ($line =~ m/^\s+(.*)$/ ) {
                $line = $1;
              }
              
              if ($line =~ m/^\s*level\s(.*)$/ ) {
                push (@recordings, [@strokes]) if ($reading);
                if ($1 eq $name) {
                  $reading = 1;
                  $records_avail++;
                  @strokes = ();
                }
                else {
                  $reading = 0;
                }
                next;
              }
              
              if ($line =~ m/^stroke\s+(.*)$/ ) {
                if ($1 eq "easy") {
                  push (@strokes, "easy");
                  next;
                }
                elsif ($1 eq "hard") {
                  push (@strokes, "hard");
                  next;
                }
                else {
                  my @stroke = ($1 =~ m/(\d+)\s*/g );
                  if ((scalar @stroke) % 3 == 0) {
                    push (@strokes, [@stroke]);
                    next;
                  }
                }    
              }

              $corrupt_record_file = $linenum if (!$corrupt_record_file);

            }

            push (@recordings, [@strokes]) if ($reading);
            
            $recording->close();
            if ($corrupt_record_file) {
              $records_avail = 0;
            }

          }
        }
        
        # Cycle through recordings and display notice.
        
        if ($playback) {
          $recording_chosen++;
          if ($recording_chosen > $records_avail) {
            $recording_chosen = 0;
            $playback = 0;
          }
        }
        elsif ($solution_avail) {
          $recording_chosen = 0;
          $playback = 1;
        }
        elsif ($records_avail) {
          $recording_chosen = 1;
          $playback = 1;
        }
        
        my $record_notice = "";
        $record_notice .= $solution_avail ? "SOLUTION AVAILABLE\n" : "NO SOLUTION AVAILABLE\n";
        if ($no_record_file) {
          $record_notice .= "CANNOT OPEN RECORD FILE\n";
        }
        elsif ($corrupt_record_file) {
          $record_notice .= "RECORD ERROR ON LINE $corrupt_record_file\n";
        }
        else {
          $record_notice .= "$records_avail RECORDING" . ($records_avail == 1 ? "" : "S") . " FOUND\n";
        }
        
        if ($playback) {
          @demo_strokes = @{$recordings[$recording_chosen]};
          if ($recording_chosen == 0) {
            $record_notice .= "SOLUTION SELECTED\n";
          }
          else {
            $record_notice .= "RECORD $recording_chosen SELECTED\n";
          }
        }
        else {
          $record_notice .= "NO RECORD SELECTED\n";
        }
        
        notice($app, $fonts[3], $record_notice, $white, $black, 320, 120);
        $app->sync();
        
      }      
    }


  }

  # Set the speed of the game.
  my $fps = 25;
  my $ticks_per_frame = 1000/$fps;
  my $frame = 0;
  my $frame_start = $app->ticks();

  # Level loop.
  while (not $end_game) {

    # Input loop.
    while (1) {
      # Check whether to switch difficulty in demo mode.
      while (($demo || $playback) && @demo_strokes && !ref($demo_strokes[0])) {
        my $diff = shift @demo_strokes;
        if ($diff eq "easy") {
          $demo_easy = 1;
        }
        elsif ($diff eq "hard") {
          $demo_easy = 0;
        }
      }

      # Check whether to fake a mouse event in demo mode.
      my $fake_mouse = 0;
      my @fake_point = ();
      my ($fake_down, $fake_up, $fake_motion) = (0, 0, 0);
      if (($demo || $playback) && @demo_strokes && ($frame == $demo_strokes[0]->[0])) {
        $fake_mouse = 1;
        my $stroke = $demo_strokes[0];
        shift (@{$stroke});
        @fake_point = splice (@{$stroke}, 0, 2);
        if (scalar @points == 0) {
          $fake_down = 1;
          if (scalar @{$stroke} == 0) {
            push (@{$stroke}, ($frame, @fake_point));
          }
        }
        elsif (scalar @{$stroke} == 0) {
          $fake_up = 1;
          shift @demo_strokes;
        }
        else {
          $fake_motion = 1;
        }
      }

      # Empty the event queue.
      last unless ($fake_mouse || $event->poll());
      my $type = 0;
      $type = $event->type() if (!$fake_mouse);
      if (!$fake_mouse && $type == SDL_QUIT) {
        # Quit immediately.
        exit(0);
      }
      elsif (!$fake_mouse && $type == SDL_KEYUP && $event->key_name() eq "escape") {
        # Restart level, or quit game entirely.
        if (!$demo) {
          $end_game = 1;
          $level_restart = 1;
        }
        else {
          exit(0);
        }
      }
      elsif (!$fake_mouse && $type == SDL_KEYUP && $event->key_name() eq "f") {
        # Toggle fullscreen.
        $app->fullscreen();
      }
      elsif (!$fake_mouse && $type == SDL_KEYUP && $event->key_name() eq "d") {
        # Toggle difficulty.
        toggle_difficulty();
        if ($demo) {
          notice($app, $fonts[3], ($easy ? "EASY" : "HARD") . " DIFFICULTY SELECTED", $white, $black);
          $app->sync();
          $app->delay(3000);
        }
      }
      elsif (!$fake_mouse && $type == SDL_KEYUP && $event->key_name() eq "s") {
        # Level skip.
        $end_game = 1;
        $level_skip = 1;
      }
      elsif ((!$fake_mouse && !($demo || $playback) && $type == SDL_MOUSEBUTTONDOWN) || ($fake_mouse && $fake_down)) {
        if ($fake_mouse || $mouse_buttons == 0) {
          # Clear previous trace.
          $draw_surf->fill($app_rect, $black);
        
          # Record first point of drawing.
          if (!$fake_mouse) {
            push (@points, [$event->button_x(), $event->button_y()]);
            if ($record) {
              $recording->print("stroke $frame " . $event->button_x() . " ". $event->button_y() . " ");
            }
          }
          else {
            push (@points, [@fake_point]);
          }
        }
        
        if ($fake_mouse) {
          $fake_mouse = 0;
        }
        else {
          $mouse_buttons = $event->button_state();
        }
      }
      elsif ((!$fake_mouse && !($demo || $playback) && $type == SDL_MOUSEBUTTONUP) || ($fake_mouse && $fake_up)) {
        $mouse_buttons = $event->button_state() if (!$fake_mouse);
        if ($mouse_buttons == 0 || $$fake_mouse) {
          # Record last point of drawing and store the whole line.
          if (!$fake_mouse) {
            push (@points, [$event->button_x(), $event->button_y()]);
            if ($record) {
              $recording->print("$frame " . $event->button_x() . " " . $event->button_y() . "\n");
            }
          } else {
            push (@points, [@fake_point]);
            $fake_mouse = 0;
          }
          push (@lines, [@points]);
          @points = ();
        }
      }
      elsif ((!$fake_mouse && !($demo || $playback) && $type == SDL_MOUSEMOTION) || ($fake_mouse && $fake_motion)) {
        if($mouse_buttons > 0 || $fake_mouse) {
          # Record mid-point of drawing and draw line segment.
          if (!$fake_mouse) {
            push (@points, [$event->motion_x(), $event->motion_y()]);
            if ($record) {
              $recording->print("$frame " . $event->button_x() . " " . $event->button_y() . " ");
            }
          }
          else {
            push (@points, [@fake_point]);
            $fake_mouse = 0;
          }
          draw_line($draw_surf, $magenta, @{$points[$#points-1]}, @{$points[$#points]});
        }
      }

      elsif ($demo && $type == SDL_MOUSEBUTTONUP) {
        # Start the game.
        $end_game = 1;
        $level_newgame = 1;
      }

    }


    # Check if any of the lines drawn are good enough to be minims.
    while (@lines) {
      my @line = @{shift @lines};
      
      # Write any change in difficulty to the recording.
      if ($recording && ($record_easy != $easy)) {
        if ($easy == 0) {
          $recording->print("stroke hard\n");
        }
        else {
          $recording->print("stroke easy\n");
        }
        $record_easy = $easy;
      }

      # First do some basic sanity checks.
      my ($x1, $y1) = @{$line[0]};
      my ($x2, $y2) = @{$line[$#line]};
      my $length = distance($x1, $y1, $x2, $y2);
      
      # Minim must be at least 10 pixels long and have 5 points.
      if ($length < 10 || @line < 5) {
        next;
      }

      # How minims are judged:
      # Only points of movement are considered; don't bother actually looking at lines between the points.
      # The first and last points become the ends of the stick of the minim.
      # So rotate and scale all the points; fit the first and last point onto the stick of an "ideal" minim.
      # The space of the ideal minim has three basic zones: the stick and a circle at both ends
      # (so that the minim can be drawn starting from either end).
      # For simplicity, each zone is represented by a rectangle.
      # The two circle zones are sub-divided into four further zones;
      # this ensures reasonable coverage of the whole circle.
      # To be accepted, all of the points drawn must lie within the stick and only one of the circles.
      # Furthermore, all five zones must have a point in them (the stick and four circle sub-zones).

      
      # Translate to put first point on (0,0).
      my @ideal = map {[$_->[0] - $x1, $_->[1] - $y1]} (@line);
      
      # Rotate and scale to put last point on (70,0).
      my $angle = atan2($ideal[$#line]->[1], $ideal[$#line]->[0]);
      my $scale = 70 / $length;
      my $cf = $scale * cos(-$angle);
      my $sf = $scale * sin(-$angle);
      @ideal = map {[($_->[0] * $cf) - ($_->[1] * $sf), ($_->[0] * $sf) + ($_->[1] * $cf)]} (@ideal);        

      # Set the tolerances for zones, based on the proportions of the drawing.
      my $stick_tol = 0;
      my $circle_tol = 0;
      my $ydev = 0;
      foreach my $point (@ideal) {
        my $y = abs($point->[1]);
        $ydev = $y if ($y > $ydev);
      }
      $ydev = 32 if ($ydev > 32);
      $stick_tol = int($ydev * 0.5);
      $circle_tol = int($ydev * 0.6);

      # Now calculate the zone boundaries, based on the tolerances.
      my ($stick_xmin, $stick_xmax, $stick_ymin, $stick_ymax) =
        (-$stick_tol, 70+$stick_tol, -$stick_tol, $stick_tol);
      my ($l_xmin, $l_xmax, $l_ymin, $l_ymax, $l_xmid, $l_ymid) =
        (-$circle_tol, $circle_tol, 2 * -$circle_tol, 0, 0, -$circle_tol);
      my ($r_xmin, $r_xmax, $r_ymin, $r_ymax, $r_xmid, $r_ymid) =
        (70-$circle_tol, 70+$circle_tol, 0, 2 * $circle_tol, 70, $circle_tol);

      
      # Check that the points lie in the right areas.

      my $fail = 0;
      # Record points that indicate it must be a top-left or bottom-right circle.
      my $left = 0;
      my $right = 0;
      # Record points falling in each circle sub-zone.
      # The stick doesn't need to be checked because the first and last points lie on it.
      my ($ltr, $ltl, $lbr, $lbl) = (0,0,0,0);
      my ($rtr, $rtl, $rbr, $rbl) = (0,0,0,0);
      foreach my $point (@ideal) {

        my ($x, $y) = @{$point};

        # Check if point is in stick area.
        if ($x >= $stick_xmin && $x <= $stick_xmax && $y >= $stick_ymin && $y <= $stick_ymax) {
          # It is; do nothing.
        }
        elsif ($x >= $l_xmin && $x <= $l_xmax && $y >= $l_ymin && $y <= $l_ymax) {
          $left++;
          $fail = 1 if $right;
        }
        elsif ($x >= $r_xmin && $x <= $r_xmax && $y >= $r_ymin && $y <= $r_ymax) {
          $right++;
          $fail = 1 if $left;
        }
        else {
          # Point is outside the allowed area.
          $fail = 1 unless (($easy && !$demo) || ($demo_easy && ($demo || $playback)));
        }
        
        # Now check sub-zones.
        $ltl = 1 if ($x >= $l_xmin && $x <= $l_xmid && $y >= $l_ymin && $y <= $l_ymid);
        $ltr = 1 if ($x >= $l_xmid && $x <= $l_xmax && $y >= $l_ymin && $y <= $l_ymid);
        $lbr = 1 if ($x >= $l_xmid && $x <= $l_xmax && $y >= $l_ymid && $y <= $l_ymax);
        $lbl = 1 if ($x >= $l_xmin && $x <= $l_xmid && $y >= $l_ymid && $y <= $l_ymax);

        $rtl = 1 if ($x >= $r_xmin && $x <= $r_xmid && $y >= $r_ymin && $y <= $r_ymid);
        $rtr = 1 if ($x >= $r_xmid && $x <= $r_xmax && $y >= $r_ymin && $y <= $r_ymid);
        $rbr = 1 if ($x >= $r_xmid && $x <= $r_xmax && $y >= $r_ymid && $y <= $r_ymax);
        $rbl = 1 if ($x >= $r_xmin && $x <= $r_xmid && $y >= $r_ymid && $y <= $r_ymax);
        
      }
      
      if ($left && $ltl && $ltr && $lbr && $lbl) {
        # It's a minim with the circle at the start.
      }
      elsif ($right && $rtl && $rtr && $rbr && $rbl) {
        # It's a minim with the circle at the end.
      }
      else {
        # Not all sub-zones contained points.
        $fail = 1;
      }
      
      # Swap the start and end points if the circle was at the end.
      if ($left > $right) {
        ($x1, $y1, $x2, $y2) = ($x2, $y2, $x1, $y1);
        $angle = atan2($y2 - $y1, $x2 - $x1);
      }

      if ($fail) {
        # If it was a failure, draw guidelines to show where the minim was expected to be.
        draw_line($draw_surf, $white, $x1, $y1, $x2, $y2);
        my $dx = int(($x2 - $x1)*($circle_tol/70));
        my $dy = int(($y2 - $y1)*($circle_tol/70));
        draw_line($draw_surf, $grey, $x2-$dy-$dx, $y2+$dx-$dy, $x2+$dx-$dy, $y2+$dx+$dy);
        draw_line($draw_surf, $white, $x2-$dy-($dx/2), $y2+$dx-($dy/2), $x2+($dx/2)-$dy, $y2+$dx+($dy/2));
        draw_line($draw_surf, $white, $x2, $y2, $x2-($dy*2), $y2+($dx*2));
      }
      else {
        # If it was a success, add the minim.
        my $dx = int(($x2 - $x1)*($circle_tol/70));
        my $dy = int(($y2 - $y1)*($circle_tol/70));
        my $x3 = $x2 - $dy;
        my $y3 = $y2 + $dx;
        my $r = int($circle_tol/$scale);

        my $minim = [$x1, $y1, $x2, $y2, $x3, $y3, $r, $length, $angle];
        draw_minim($note_surf, $cyan, @{$minim});
        push (@minims, $minim);
        
        # Log all minims drawn and the frame and level number to the "log" file in the current directory.
        while(0)
        {
          my $log = new IO::File("log", ">>");
          $log->print("$stage $frame: $x1 $y1 $x2 $y2 $r\n");
          $log->close();
        }
      }
      
    }
    
    # Update game state.

    # Time-related events.
    
    # Check for second boundary.
    if (($frame % $fps) == 0) {
      $seconds++;

      # Update on-screen clock.
      font_print($fonts[3], $bg_surf, sprintf("%3d", $time_limit - $seconds), 565, 460);

      if (($time_limit - $seconds) == 5) {
        &$play("timenear") if (!$demo);
      }
      
      # Check if the time has run out.
      if ($seconds >= $time_limit) {
        &$play("timeout") if (!$demo);
        $end_game = 1;
        $level_timeover = 1;
      }

      # Check for streams that become active.
      push (@active_streams, map {[0, @{$_}]} (grep {$_->[4] == $seconds} (@streams)));

      # Remove streams that become inactive.
      @active_streams = grep {$_->[6] > $seconds} (@active_streams);

    }

    # Add new semibreves if necessary.
    foreach my $stream (@active_streams) {
      if($stream->[0] <= 0) {
        push (@semibreves, [$stream->[1], $stream->[2], $stream->[3], $stream->[4], $stream->[8]]);
        $stream->[0] = $stream->[7];
      }
      $stream->[0]--;
    }

    # Increment frame counter.
    $frame++;
    
    # Move the semibreves.
    my $del_minim = 0;
    my $del_semi = 0;
    foreach my $semi (@semibreves) {
      my ($x, $y, $velx, $vely, $colour) = @{$semi};

      my $newx = $x + $velx;
      my $newy = $y + $vely;
      # Gravity.
      my $grav = 1;

      # Check for reaching the exit.
      foreach my $exit (@exits) {
        if (defined($colour) && ($colour == $exit->[2])) {
          my ($x1, $x2, $y1, $y2) = ($exit->[0] - 10, $exit->[0] + 10, $exit->[1] - 10, $exit->[1] + 10);
          if ($x >= $x1 && $x <= $x2 && $y >= $y1 && $y <= $y2) {
            if ($targets[$colour] > 0) {
              &$play("exit") if (!$demo);
              $targets[$colour]--;
              font_print($fonts[$colour], $bg_surf, sprintf("%03d", $targets[$colour]), 185 + (85*$colour), 460);
            }
            if (!$demo && $targets[0] == 0 && $targets[1] == 0 && $targets[2] == 0) {
              $end_game = 1;
              $level_success = 1;
            }
            $colour = undef;
            $del_semi = 1;
            last;
          }
        }
      }

      # Check for leaving the screen.
      if ($x < -10 || $y < -10 || $x > 650 || $y > 490) {
        $colour = undef;
        $del_semi = 1;
      }

      # Check for collision with rests.
      foreach my $sniff (@rests) {
        if (defined($colour)) {
          my ($x1, $x2, $y1, $y2) = ($sniff->[0] - 10, $sniff->[0] + 10, $sniff->[1] - 10, $sniff->[1] + 10);
          if ($x >= $x1 && $x <= $x2 && $y >= $y1 && $y <= $y2) {
            # Slow down.
            $velx *= $slow;
            $vely *= $slow;
          }
        }
      }

      
      # Check for collision with minims.
      foreach my $minim (@minims) {
        next if !defined($minim);

        my ($x1, $y1, $x2, $y2, $x3, $y3, $r, $length, $angle) = @{$minim};
        
        # Give the semibreve a boost when near the circle.
        if (($r > 0) && distance($newx, $newy, $x3, $y3) < ($r + $semi_rad)) {
          $velx *= $boost;
          $vely *= $boost;
        }
         
        # Check for crossing.
        my ($tx1, $tx2) = ($x - $x1, $newx - $x1);
        my ($ty1, $ty2) = ($y - $y1, $newy - $y1);
        my $cf = cos(-$angle);
        my $sf = sin(-$angle);
        ($tx1, $ty1, $tx2, $ty2) =
          map {(($_->[0] * $cf) - ($_->[1] * $sf), ($_->[0] * $sf) + ($_->[1] * $cf))}
          ([$tx1, $ty1], [$tx2, $ty2]);

        if ($tx1 >= -$semi_rad && $tx1 <= ($length + $semi_rad) &&
          $tx2 >= -$semi_rad && $tx2 <= ($length + $semi_rad)) {

          if ($r > 0 && $ty1 < -$semi_rad && $ty2 >= -$semi_rad) {
            # Collision from behind.
            # Destroy semibreve and minim.
            &$play("break") if (!$demo);  
            $redraw_note = 1;
            $minim = undef;
            $del_minim = 1;
            $colour = undef;
            $del_semi = 1;
            last;
          }
          elsif ($ty1 > $semi_rad && $ty2 <= $semi_rad) {
            # Collision from in front.
            # Bounce semibreve.
            &$play("bounce") if (!$demo);
            $grav = 0;
            ($newx, $newy) = ($x, $y);
            ($velx, $vely) =
              map {(($_->[0] * $cf) - ($_->[1] * $sf), ($_->[0] * $sf) + ($_->[1] * $cf))} ([$velx, $vely]);
            $vely *= -$elasticity;
            $cf = cos($angle);
            $sf = sin($angle);
            ($velx, $vely) =
              map {(($_->[0] * $cf) - ($_->[1] * $sf), ($_->[0] * $sf) + ($_->[1] * $cf))} ([$velx, $vely]);

          }
          
        }

      }

      $semi = [$newx, $newy, $velx + ($grav ? $gravity[0] : 0), $vely + ($grav ? $gravity[1] : 0), $colour];

    }

    # Actually remove any minims or semibreves that were destroyed.

    if ($del_minim) {
      for (my $n = $#minims; $n >= 0; $n--) {
        if (!defined($minims[$n])) {
          splice @minims, $n, 1;
        }
      }
      $del_minim = 0;
    }

    if ($del_semi) {
      for (my $n = $#semibreves; $n >= 0; $n--) {
        if (!defined($semibreves[$n]->[4])) {
          splice @semibreves, $n, 1;
        }
      }
      $del_semi = 0;
    }
    

    # Update display.
    $bg_surf->blit($app_rect, $app, $app_rect);
    if ($redraw_note) {
      $note_surf->fill($app_rect, $black);
      foreach my $minim (@minims) {
        draw_minim($note_surf, $cyan, @{$minim});
      }
      $redraw_note = 0;
    }
    $note_surf->blit($app_rect, $app, $app_rect);
    foreach my $semi (@semibreves) {
      my ($x, $y, $velx, $vely, $colour) = @{$semi};
      #draw_circle($app, $colour, $x, $y, $semi_rad);
      $semi_targ->x($x - $semi_rad);
      $semi_targ->y($y - $semi_rad);
      $semis[$colour]->blit($semi_rect, $app, $semi_targ);
    }
    $draw_surf->blit($app_rect, $app, $app_rect);
    $app->sync();
    # Save a string of numbered screenshots in the current directory.
    #$app->save_bmp(sprintf("%08d", $frame) . ".bmp");
    
    # Wait for next frame if necessary.
    {
      my $now = $app->ticks();
      my $diff = $now - $frame_start;
      if ($diff < $ticks_per_frame) {
        $app->delay($ticks_per_frame - $diff);
        $frame_start += $ticks_per_frame;
      }
      else {
        $frame_start = $now;
      }
    }

  }

  # Check how the level ended.
  if ($demo) {
    if ($level_skip || $level_timeover) {
      # Level skip or end of demo level.
      $stage++;
      if ($stage >= @demo_levels) {
        $stage = 0;
      }
    }
    else {    
      # Start game.
      $stage = 0;
      $demo = 0;
      &$play("click") if (!$demo);
    }
  }
  
  else {
    my $text = "";

    if ($level_skip) {
      # Level skip.
      &$play("click") if (!$demo);
      $stage++;
      $text = "LEVEL SKIPPED";
    }
    elsif ($level_restart) {
      # Level aborted.
      &$play("click") if (!$demo);
      $text = "LEVEL ABORTED";
    }
    elsif ($level_timeover && ($targets[0]>0 || $targets[1]>0 || $targets[2]>0)) {
      # Level failed.
      $text = "OUT OF TIME";
    }
    else {
      # Level passed.
      &$play("clear") if (!$demo);
      # Do not advance if the level was completed by playback.
      $stage++ unless $playback;
      $text = "LEVEL COMPLETE";
    }
    
    # Stop recording.
    if ($record) {
      $recording->print("# Stopping recording: $text\n");
      $recording->close();
    }
    
    # Return to demo if game complete.
    if ($stage >= @game_levels) {
      $demo = 1;
      $stage = 0;
    }

    # Display a notice.
    notice($app, $fonts[3], $text, $white, $black);
    $app->sync();

    # Wait for click.
    while ($event->wait()) {
      my $type = $event->type();
      if ($type == SDL_QUIT) {
        # Quit immediately.
        exit(0);
      }
      elsif ($type == SDL_MOUSEBUTTONUP) {
        # Proceed to start of level.
        &$play("click") if (!$demo);
        last;
      }
      elsif ($type == SDL_KEYUP && $event->key_name() eq "f") {
        # Toggle fullscreen.
        $app->fullscreen();
      }
      elsif ($type == SDL_KEYUP && $event->key_name() eq "d") {
        # Toggle difficulty.
        toggle_difficulty();
      }
      # Ignore escape to avoid accidental quits.
      
    }

  }

}

